-- name=dbg_safe_api
--------------------------------------------------------------------------------
-- NotEngine API extensions.
-- Reimplementation of some NotEngine API functions with verbose traces
-- @module dbg_safe_api
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Setup NotEngine API changes.
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  --- Enable some additional strict checks.
  private.pedantic = true;
  
  --- Enable DB query trace.
  private.trace_queries = false;
  
  --- Trace DB queries type.
  private.trace_queries_type = false;
  
  --- Trace DB queries and result.
  private.trace_queries_data = true;
  
  --- Trace DB queries time.
  private.trace_queries_time = false;
  
  --- Trace DB queries total count and time.
  private.trace_queries_total = false;
  private.queries_count = 0;
  private.queries_time  = 0;

  -- Enable tracing queries filters ( string.match patterns )
  private.trace_queries_filter = false;
  private.trace_queries_filter_type = '.*';
  private.trace_queries_filter_collection = 'player/stages.*';
  
  -- Trace depth shifter, see private.GetErrorSource, private.TraceDepthMore, 
  -- private.TraceDepthLess
  private.tarce_depth_shift = 0;
  
  private.SetupStrictData();
  
  local function redefine ( list, func_strict )
    private.Replace( list, "__Raw", func_strict );
  end;
  
  redefine( { "ne.dbg.Trace",     "DbgTrace"    }, private.DbgTrace    );
  redefine( { "ne.obj.Create",    "ObjCreate"   }, private.ObjCreate   );
  redefine( { "ne.obj.Attach",    "ObjAttach"   }, private.ObjAttach   );
  redefine( { "ne.obj.Set",       "ObjSet"      }, private.ObjSet      );
  redefine( { "ne.obj.MultiSet",  "ObjMultiSet" }, private.ObjMultiSet );
  redefine( { "ne.obj.Get",       "ObjGet"      }, private.ObjGet      );
  redefine( { "ne.obj.anim.Play", "AnimPlay"    }, private.AnimPlay    );
  redefine( { "ne.obj.Animate",   "ObjAnimate"  }, private.ObjAnimate  );
  redefine( { "ne.obj.Clone",     "ObjClone"    }, private.ObjClone    );
  redefine( { "GetObjPosByObj"                  }, private.ObjPosByObj );
  
  -- Debug data for ModLoad
  _PRIVATE = {};
  if ( _G[ "loader" ] and _G[ "loader" ].__Private ) then
  
    _PRIVATE[ "loader" ] = _G[ "loader" ].__Private();
  
  end;
  
  redefine( { "ne.mod.Load",      "ModLoad" }, private.ModLoad );
  redefine( { "ne.mod.LoadLazily"           }, private.ModLoad );
  
  if ( private.trace_queries ) then
  
    redefine( { "DbExecuteQuery" }, private.DbExecuteQuery );
  
  end;
  
end;

--------------------------------------------------------------------------------
-- Setup data tables for strict checks processing.
--------------------------------------------------------------------------------
function private.SetupStrictData ()
  
  -- Object types list
  private.object_types = {
    ["anim"]              = true,
    ["dsprite"]           = true,
    ["grid"]              = true,
    ["matchbox"]          = true,
    ["obj"]               = true,
    ["partsys"]           = true,
    ["scrollcontainer"]   = true,
    ["spr"]               = true,
    ["text"]              = true,
    ["textadapt"]         = true,
    ["textedit"]          = true,
    ["timer"]             = true,
    ["transparentobject"] = true,
    ["video"]             = true,
    ["zoomview"]          = true
  }
  
  -- generic object parameters
  private.checkers     = {};
  private.checkers.obj = {
    ["name"] = private.IsStr,
    ["res"]  = private.IsStr,

    ["active"]  = private.CheckParam_Flag,
    ["visible"] = private.CheckParam_Flag,
    ["input"]   = private.CheckParam_Flag,

    ["grayscale"]  = private.CheckParam_Flag,
    ["alp"]        = private.CheckParam_Fraction,
    ["color_r"]    = private.CheckParam_Fraction,
    ["color_g"]    = private.CheckParam_Fraction,
    ["color_b"]    = private.CheckParam_Fraction,

    ["ang"]     = private.IsNum,
    ["pos_x"]   = private.IsNum,
    ["pos_y"]   = private.IsNum,
    ["pos_z"]   = private.IsNum,
    ["scale_x"] = private.IsNum,
    ["scale_y"] = private.IsNum,

    ["drag"]     = private.CheckParam_Flag,
    ["realdrag"] = private.CheckParam_Flag,

    ["inputrect_init"] = private.CheckParam_Flag,
    ["inputrect_x"]    = private.IsNum,
    ["inputrect_y"]    = private.IsNum,
    ["inputrect_w"]    = private.CheckParam_PosNumber,
    ["inputrect_h"]    = private.CheckParam_PosNumber,

    ["croprect_init"] = private.CheckParam_Flag,
    ["croprect_x"]    = private.IsNum,
    ["croprect_y"]    = private.IsNum,
    ["croprect_w"]    = private.CheckParam_PosNumber,
    ["croprect_h"]    = private.CheckParam_PosNumber,

    ["draw_height"] = private.CheckParam_PosNumber,
    ["draw_width"]  = private.CheckParam_PosNumber,
    ["drawoff_x"]   = private.IsNum,
    ["drawoff_y"]   = private.IsNum,

    ["attachtype"]  = private.CheckParam_AttachType,
    ["blendmode"]   = private.CheckParam_BlendMode,
    ["blur_radius"] = private.CheckParam_PosNumber,
    ["anim_tag"]    = private.IsStr,

    ["event_mdown"]       = private.CheckParam_Handler,
    ["event_mup"]         = private.CheckParam_Handler,
    ["event_menter"]      = private.CheckParam_Handler,
    ["event_mleave"]      = private.CheckParam_Handler,
    ["event_menter_drag"] = private.CheckParam_Handler,
    ["event_mleave_drag"] = private.CheckParam_Handler,
    ["event_dragdrop"]    = private.CheckParam_Handler,
    ["event_drag"]        = private.CheckParam_Handler,
    ["event_startdrag"]   = private.CheckParam_Handler
  };
  
  -- transparent object
  private.checkers.transparentobject = private.checkers.obj;
  
  -- partsys
  private.checkers.partsys = private.checkers.obj;  
  
  -- sprite parameters
  private.checkers.spr = {
    ["frame"]         = private.CheckParam_PosNumber,
  };
  private.ExtendTable( private.checkers.spr, private.checkers.obj );
  
  -- dsprite parameters
  private.checkers.dsprite = {
    ["trg_resloaded"] = private.CheckParam_Handler,
    ["default_res"]   = private.IsStr,
    ["url"]           = private.IsStr
  };
  private.ExtendTable( private.checkers.dsprite, private.checkers.spr );
  
  -- grid parameters
  private.checkers.grid = {
    ["playing"]  = private.CheckParam_Flag,
    ["endtrig"]  = private.CheckParam_Handler
  };
  private.ExtendTable( private.checkers.grid, private.checkers.spr );
  
  -- anim parameters
  private.checkers.anim = {
    ["animfunc"] = private.IsStr
  };
  private.ExtendTable( private.checkers.anim, private.checkers.grid );
  
  -- video parameters
  private.checkers.video = {
    ["looped"] = private.CheckParam_Flag
  };
  private.ExtendTable( private.checkers.video, private.checkers.obj );
  
  -- timer parameters
  private.checkers.timer = {
    ["endtrig"]  = private.CheckParam_Handler,
    ["playing"]  = private.CheckParam_Flag,
    ["infinite"] = private.CheckParam_Flag,
    ["time"]     = private.CheckParam_PosNumber
  };
  private.ExtendTable( private.checkers.timer, private.checkers.obj );
  
  -- zoomview parameters
  private.checkers.zoomview = {
    ["candrag"]        = private.CheckParam_Flag,
    ["canzoom"]        = private.CheckParam_Flag,
    ["canusevelocity"] = private.CheckParam_Flag,

    ["scale"]          = private.CheckParam_PosNumber,
    ["max_scale"]      = private.CheckParam_PosNumber,
    ["min_scale"]      = private.CheckParam_PosNumber,

    ["scrollrect_x"]   = private.IsNum,
    ["scrollrect_y"]   = private.IsNum,
    ["scrollrect_w"]   = private.CheckParam_PosNumber,
    ["scrollrect_h"]   = private.CheckParam_PosNumber,

    ["screen_x"]       = private.IsNum,
    ["screen_y"]       = private.IsNum,
    ["screen_w"]       = private.CheckParam_PosNumber,
    ["screen_h"]       = private.CheckParam_PosNumber,

    ["begindragtrig"]  = private.CheckParam_Handler,
    ["enddragtrig"]    = private.CheckParam_Handler,
    ["dragtrig"]       = private.CheckParam_Handler
  };
  private.ExtendTable( private.checkers.zoomview, private.checkers.obj );
  
  -- scrollcontainer parameters
  private.checkers.scrollcontainer = {
    ["horizontal"]     = private.CheckParam_Flag,
    ["dragenabled"]    = private.CheckParam_Flag,
    ["cellsize"]       = private.CheckParam_PosNumber,
    ["scrollpos"]      = private.CheckParam_PosNumber,
    ["begindragtrig"]  = private.CheckParam_Handler,
    ["enddragtrig"]    = private.CheckParam_Handler,
    ["scrollbacktrig"] = private.CheckParam_Handler,
    ["scrollendtrig"]  = private.CheckParam_Handler
  };
  private.ExtendTable( private.checkers.scrollcontainer, private.checkers.obj );
  
  -- matchbox parameters
  private.checkers.matchbox = {
    ["ap"]              = private.CheckParam_Flag,
    ["stage"]           = private.IsStr,
    ["state"]           = private.CheckParam_PosNumber,
    ["section"]         = private.CheckParam_PosNumber,
    ["attempts"]        = private.CheckParam_PosNumber,
    ["max_attempts"]    = private.CheckParam_PosNumber,
    ["preset"]          = private.CheckParam_Flag,
    ["boosters"]        = private.IsTab,
    ["no_limits_value"] = private.CheckParam_PosNumber
  };
  private.ExtendTable( private.checkers.matchbox, private.checkers.obj );
  
  -- text parameters
  private.checkers.text = {
    ["text"]        = private.CheckParam_StringParam,
    ["disprawtext"] = private.CheckParam_Flag,

    ["fontsize"] = private.CheckParam_PosNumber,
    ["fonttype"] = private.CheckParam_PosNumber,

    ["linespacing"] = private.CheckParam_PosNumber,
    ["maxwidth"]    = private.CheckParam_PosNumber,

    ["align"]   = private.CheckParam_TextAlign,
    ["align_v"] = private.CheckParam_TextAlign,

    ["fontcolor_r"] = private.CheckParam_Fraction,
    ["fontcolor_g"] = private.CheckParam_Fraction,
    ["fontcolor_b"] = private.CheckParam_Fraction,

    ["display_outline"] = private.CheckParam_Flag,
    ["outline_color_r"] = private.CheckParam_Fraction,
    ["outline_color_g"] = private.CheckParam_Fraction,
    ["outline_color_b"] = private.CheckParam_Fraction,
    ["outline_alpha"]   = private.CheckParam_Fraction,
    ["outline_size"]    = private.CheckParam_PosNumber,

    ["display_shadow"] = private.CheckParam_Flag,
    ["shadow_color_r"] = private.CheckParam_Fraction,
    ["shadow_color_g"] = private.CheckParam_Fraction,
    ["shadow_color_b"] = private.CheckParam_Fraction,
    ["shadow_aplha"]   = private.CheckParam_Fraction,
    ["shadow_pos_x"]   = private.IsNum,
    ["shadow_pos_y"]   = private.IsNum,

    ["param0"] = private.CheckParam_StringParam,
    ["param1"] = private.CheckParam_StringParam,
    ["param2"] = private.CheckParam_StringParam,
    ["param3"] = private.CheckParam_StringParam,
    ["param4"] = private.CheckParam_StringParam
  };
  private.ExtendTable( private.checkers.text, private.checkers.obj );
  
  -- textadapt parameters
  private.checkers.textadapt = {
    ["shorttext_type"] = private.CheckParam_ShortText,
    ["min_fontsize"]   = private.CheckParam_PosNumber,
    ["max_fontsize"]   = private.CheckParam_PosNumber,
    ["min_scale_x"]    = private.CheckParam_PosNumber,
    ["delay_letter"]   = private.CheckParam_PosNumber
  };
  private.ExtendTable( private.checkers.textadapt, private.checkers.text );
  
  -- textedit parameters
  private.checkers.textedit = {
    ["allowed_characters"] = private.IsStr,
    ["maxchars"]           = private.CheckParam_PosNumber,
    ["clearonfirstinput"]  = private.CheckParam_Flag,
    ["trg_enter"]          = private.CheckParam_Handler,
    ["trg_onfirstinput"]   = private.CheckParam_Handler,
    ["show_native"]        = private.CheckParam_Flag,
    ["cursorpos"]          = private.CheckParam_PosNumber,
    ["cursorcolor_r"]      = private.CheckParam_Fraction,
    ["cursorcolor_g"]      = private.CheckParam_Fraction,
    ["cursorcolor_b"]      = private.CheckParam_Fraction,
    ["native_h"]           = private.CheckParam_PosNumber,
    ["native_w"]           = private.CheckParam_PosNumber,
    ["native_y"]           = private.IsNum,
    ["native_x"]           = private.IsNum
  };
  private.ExtendTable( private.checkers.textedit, private.checkers.text );
  
  -- Non custom animation checkers, see private.ObjAnimate_CheckAnimTable
  private.non_custom_anims = {
    ["pos_xy"]    = { 2, private.IsNum },
    ["scale_xy"]  = { 2, private.IsNum },
    ["color_rgb"] = { 3, private.CheckParam_Fraction }
  };
  
end;

--------------------------------------------------------------------------------
-- Add 'base_tab' keys to 'tab' table.
-- @param tab extending table.
-- @param base_tab base table.
--------------------------------------------------------------------------------
function private.ExtendTable( tab, base_tab )

  for k,v in pairs( base_tab ) do
    
    if not ( tab[k] ) then tab[k] = v; end;
    
  end;
    
end;

--------------------------------------------------------------------------------
-- function ... () [ SPACER ]
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Split string by '.' separator.
-- @param str string for split
-- @return list of substrings
--------------------------------------------------------------------------------
function private.Split ( str )

  local result = {};
  
  for token in str:gmatch("([^%.]+)") do
    table.insert(result, token)
  end
  
  return result;

end;

--------------------------------------------------------------------------------
-- function ... () [ SPACER ]
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Replace functions by other.
-- Old function marked by prefix
-- @param list list of function names
-- @param prefix for renaming old functions
-- @param func_repalce pointer to replacing function
--------------------------------------------------------------------------------
function private.Replace ( list, prefix, func_repalce )

  for _, str in pairs( list ) do
  
    local parent = _G;
    local tokens = private.Split( str );
    
    for i, token in pairs( tokens ) do
    
      if ( i == #tokens ) then
      
        local func_orig = parent[ token ];
        parent[ prefix..token ] = func_orig;
      
        parent[ token ] = function ( ... )

          return func_repalce( func_orig, ... );
        
        end;
        
      else
      
        parent = parent[ token ];
      
      end
    
    end;
  
  end;
  
end;

--------------------------------------------------------------------------------
-- function ... () [ SPACER ]
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Get function call source in stack.
-- @return string in format 'module_name:sting_number'.
--------------------------------------------------------------------------------
function private.GetSource ()
  
  local source = "unknown";
  local strnum = -1;
  
  source = source..":"..strnum;
  
  -- Get error source if debug mode enabled
  if ( debug ) then
    
    -- debug.getinfo() parameters:
    local level  = 5;     -- 0 = debug.getinfo,
                          -- 1 = private.GetSource,
                          -- 2 = private.Trace,
                          -- 3 = private.TraceError, private.TraceInfo
                          -- 4 = safe API func
                          -- 5 = func, where safe func called
                        
    local fields = "Sl";  -- "S" select fields { source, ... }
                          -- "l" select field  { currentline }

    level = level + private.tarce_depth_shift;

    local info = debug.getinfo( level, fields );
    
    if ( info ) then
    
      source = info.source..":"..info.currentline;
    
      info = debug.getinfo( level + 1, fields );
      
      if ( info ) then
    
        source = info.source..":"..info.currentline.." -> "..source;
      
      end;
        
    end;
    
  end;
  
  return source;
  
end;

--------------------------------------------------------------------------------
-- Generic tracing function. Use in 'strict' mode.
-- @param cat tarce category, like 'ERROR', 'WARNING', 'INFO'.
-- @param who trace caller, function name.
-- @param msg error message.
-- @param[opt] tracer tracing function (DbgTrace default).
--------------------------------------------------------------------------------
function private.Trace ( cat, who, msg, tracer )
  
  local source = private.GetSource();
  local trace = "[ SAFE API ][ " .. cat .. " ] " 
                .. source .. " " .. who .. ": " .. msg;
  
  tracer = tracer or DbgTrace;
  tracer( trace );
  
end;

--------------------------------------------------------------------------------
-- Error tracing function.
-- @param who trace caller, function name.
-- @param msg error message.
-- @param[opt] tracer tracing function (DbgTrace default).
--------------------------------------------------------------------------------
function private.TraceError ( who, msg, tracer )
  
  private.Trace( "ERROR", who, msg, tracer );
  
end;

--------------------------------------------------------------------------------
-- Warning tracing function.
-- @param who trace caller, function name.
-- @param msg warning message.
-- @param[opt] tracer tracing function (DbgTrace default).
--------------------------------------------------------------------------------
function private.TraceWarning ( who, msg, tracer )
  
  private.Trace( "WARNING", who, msg, tracer );
  
end;

--------------------------------------------------------------------------------
-- Info tracing function.
-- @param who trace caller, function name.
-- @param msg info message.
-- @param[opt] tracer tracing function (DbgTrace default).
--------------------------------------------------------------------------------
function private.TraceInfo ( who, msg, tracer )
  
  private.Trace( "INFO", who, msg, tracer );
  
end;

--------------------------------------------------------------------------------
-- Bad type error tracing function.
-- @param who trace caller, function name.
-- @param key parameter name.
-- @param value parameter value.
--------------------------------------------------------------------------------
function private.TraceBadType ( who, key, value )
  
  private.TraceDepthMore();
  
  private.TraceError( who, "bad " .. key .. " type: " 
                           .. type( value )
                           .. " => " 
                           .. private.AnyToStr( value ) );

  private.TraceDepthLess();
  
end

--------------------------------------------------------------------------------
-- Make error source more deep in stack.
--------------------------------------------------------------------------------
function private.TraceDepthMore ()
  
  private.tarce_depth_shift = private.tarce_depth_shift + 1;
  
end;

--------------------------------------------------------------------------------
-- Make error source less deep in stack.
--------------------------------------------------------------------------------
function private.TraceDepthLess ()
  
  private.tarce_depth_shift = private.tarce_depth_shift - 1;
  
end;

--------------------------------------------------------------------------------
-- function ... () [ SPACER ]
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Check if value is boolean.
-- @param value value for check.
-- @return true if value type is bool.
--------------------------------------------------------------------------------
function private.IsBool ( value )
  
  return type( value ) == "boolean";
  
end;

--------------------------------------------------------------------------------
-- Check if value is number.
-- @param value value for check.
-- @return true if value type is bool.
--------------------------------------------------------------------------------
function private.IsNum ( value )
  
  return type( value ) == "number";
  
end;

--------------------------------------------------------------------------------
-- Check if value is string.
-- @param value value for check.
-- @return true if value type is string.
--------------------------------------------------------------------------------
function private.IsStr ( value )
  
  return type( value ) == "string";
  
end;

--------------------------------------------------------------------------------
-- Check if value is table.
-- @param value value for check.
-- @return true if value type is table.
--------------------------------------------------------------------------------
function private.IsTab ( value )
  
  return type( value ) == "table";
  
end;

--------------------------------------------------------------------------------
-- Check if value is function.
-- @param value value for check.
-- @return true if value type is function.
--------------------------------------------------------------------------------
function private.IsFunc ( value )
  
  return type( value ) == "function";
  
end;

--------------------------------------------------------------------------------
-- function ... () [ SPACER ]
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Check if object exists.
-- @param name object name.
-- @return true if object exists.
--------------------------------------------------------------------------------
function private.IsObjExists ( name )
  
  if ( ObjGet( name ) ) then
    return true;
  else
    return false;
  end;
  
end;

--------------------------------------------------------------------------------
-- Check if object is tagged type (animation).
-- @param obj_name object name.
-- @return true if object is animation.
--------------------------------------------------------------------------------
function private.IsTaggedObject ( obj_name )

  return ( sml.obj.GetType( obj_name ) == "anim" );
  
end;

--------------------------------------------------------------------------------
-- Check if value is integer.
-- @param value value for check.
-- @return true if value type is integer.
--------------------------------------------------------------------------------
function private.CheckParam_Integer( value )
  
  return private.IsNum( value ) and ( ( value % 1 ) == 0 );
  
end;

--------------------------------------------------------------------------------
-- Check if number in range include bounds.
-- @param min minimum value.
-- @param value value for check.
-- @param max maximum value.
-- @return true if value in range.
--------------------------------------------------------------------------------
function private.CheckParam_InRange ( min, value, max )
  
  return private.IsNum( value ) and ( value >= min ) and ( value <= max );
  
end;

--------------------------------------------------------------------------------
-- Check if value in [0, 1].
-- @param value value for check.
-- @return true if value in range.
--------------------------------------------------------------------------------
function private.CheckParam_Fraction ( value )
  
  return private.CheckParam_InRange( 0, value, 1 );
  
end;

--------------------------------------------------------------------------------
-- Check if value is valid blend mode.
-- @param value value for check.
-- @return true if value is valid blend mode (see wiki).
--------------------------------------------------------------------------------
function private.CheckParam_BlendMode ( value )
  
  return private.CheckParam_Integer( value )
     and private.CheckParam_InRange( 0, value, 4 );
  
end;

--------------------------------------------------------------------------------
-- Check if value is valid attach type.
-- @param value value for check.
-- @return true if value is valid attach type (see wiki).
--------------------------------------------------------------------------------
function private.CheckParam_AttachType ( value )
  
  return private.CheckParam_Integer( value )
     and private.CheckParam_InRange( 0, value, 3 );
  
end;

--------------------------------------------------------------------------------
-- Check if value is valid text align.
-- @param value value for check.
-- @return true if value is valid text align (see wiki).
--------------------------------------------------------------------------------
function private.CheckParam_TextAlign ( value )
  
  return private.CheckParam_Integer( value )
     and private.CheckParam_InRange( 0, value, 2 );
  
end;

--------------------------------------------------------------------------------
-- Check if value is valid short text type.
-- @param value value for check.
-- @return true if value is valid short text type (see wiki).
--------------------------------------------------------------------------------
function private.CheckParam_ShortText ( value )
  
  return private.CheckParam_Integer( value )
     and private.CheckParam_InRange( 0, value, 1 );
  
end;

--------------------------------------------------------------------------------
-- Check if value is flag (boolean or 1/0).
-- @param value value for check.
-- @return true if value is flag.
--------------------------------------------------------------------------------
function private.CheckParam_Flag ( value )
  
  return private.IsBool( value ) or ( value == 0 ) or ( value == 1 );
  
end;

--------------------------------------------------------------------------------
-- Check if value is handler (function or string).
-- @param value value for check.
-- @return true if value is handler.
--------------------------------------------------------------------------------
function private.CheckParam_Handler ( value )
  
  return private.IsStr( value ) or private.IsFunc( value ) ;
  
end;

--------------------------------------------------------------------------------
-- Check if value is string parameter (number or string).
-- @param value value for check.
-- @return true if value is string parameter.
--------------------------------------------------------------------------------
function private.CheckParam_StringParam ( value )
  
  return private.IsNum( value ) or private.IsStr( value ) ;
  
end;

--------------------------------------------------------------------------------
-- Check if value is positive number.
-- @param value value for check.
-- @return true if value is positive number.
--------------------------------------------------------------------------------
function private.CheckParam_PosNumber ( value )
  
  return private.IsNum( value ) and ( value >= 0 ) ;
  
end;

--------------------------------------------------------------------------------
-- function ... () [ SPACER ]
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Strict version of DbgTrace.
-- @param func_orig original function.
-- @param trace tracing text (string).
--------------------------------------------------------------------------------
function private.DbgTrace ( func_orig, trace )
  
  if not ( private.IsStr( trace ) ) then
    
    local who = "DbgTrace";
    local t   = type( trace );
    
    private.TraceError( who, "try to trace non-string: " .. t, func_orig );
    
  end;
  
  return func_orig( trace );
  
end;

--------------------------------------------------------------------------------
-- Strict version of ObjCreate.
-- @param func_orig original function.
-- @param name object name, string.
-- @param obj_type object type, string.
--------------------------------------------------------------------------------
function private.ObjCreate ( func_orig, name, obj_type )
  
  local who = "ObjCreate";

  -- Check 'name'
  if not ( private.IsStr( name ) ) then
    
    private.TraceBadType( who, "object name", name );
    
  elseif ( private.IsObjExists( name ) ) then
    
    private.TraceError( who, "try to create exists object: '" .. name .. "'" );
    
  end;
  
  -- Check 'obj_type'
  if not ( private.IsStr( obj_type ) ) then
    
    private.TraceBadType( who, "object type name", obj_type );
    
  elseif not ( private.object_types[obj_type] ) then
    
    private.TraceError( who, "unknown object type: " .. obj_type );
    
  end;
  
  -- Pass data to original API
  return func_orig( name, obj_type );
  
end;

--------------------------------------------------------------------------------
-- Strict version of ObjAttach.
-- @param func_orig original function.
-- @param object object name, string.
-- @param parent object parent, string.
--------------------------------------------------------------------------------
function private.ObjAttach ( func_orig, object, parent )
  
  local who = "ObjAttach";
  
  -- Check 'object' argument
  if not ( private.IsStr( object ) ) then
    
    private.TraceBadType( who, "object name", object );
    
  elseif not ( private.IsObjExists( object ) ) then
    
    private.TraceError( who, "object does not exist: '" .. object .. "'" );
    
  end;
  
  -- Check 'parent' argument
  if not ( private.IsStr( parent ) ) then
    
    private.TraceBadType( who, "parent name", parent );
    
  elseif not ( private.IsObjExists( parent ) ) then
    
    private.TraceError( who, "parent does not exist: '" .. parent .. "'" );
    
  end;
  
  -- Pass data to original API
  return func_orig( object, parent );
  
end;

--------------------------------------------------------------------------------
-- Strict version of ObjMultiset.
-- @param func_orig original function.
-- @param tab objects and params table.
--------------------------------------------------------------------------------
function private.ObjMultiSet ( func_orig, tab )
  
  local who = "ObjMultiset";
  
  if not ( private.IsTab( tab ) ) then
    
    private.TraceError( who, "invalid argument type: " .. type( object ) );
    
  elseif not ( #tab > 0 ) then
    
    private.TraceError( who, "set table is empty" );
    
  else
    
    for id = 1, #tab do
      
      local entry    = tab[id];
      local obj_name = entry[1];
      local params   = entry[2];
      
      if not ( obj_name and params ) then
        
        private.TraceError( who, "invalid set entry data:" ..
                                 " object name:"  .. tostring( obj_name ) ..
                                 ", params:"      .. tostring( params ) );
        
      elseif not ( private.IsStr( obj_name ) and private.IsTab( params ) ) then
        
        private.TraceError( who, "invalid set entry types:" ..
                                 " object name:"  .. type( obj_name ) ..
                                 ", params:"      .. type( params ) );

      else

        if not ( private.IsObjExists( obj_name ) ) then
          
          private.TraceError( who, "object does not exists " .. "'" ..
                                    obj_name .. "'" );
          
        end;
        
        if ( private.pedantic ) then
          
          private.ObjSet_CheckParams( obj_name, params, who );
          
        end;
        
      end;
      
    end;
    
  end;
  
  return func_orig( tab );
  
end;

--------------------------------------------------------------------------------
-- Strict version of ObjGet.
-- @param func_orig original function.
-- @param object object name, string.
-- @return pbject parameters table (see wiki).
--------------------------------------------------------------------------------
function private.ObjGet ( func_orig, object )
  
  local who = "ObjGet";
  
  -- Check 'object' type
  if not ( private.IsStr( object ) ) then
    
    private.TraceBadType( who, "object name", object );
    
  end;
  
  -- Pass data to original API
  return func_orig( object );
  
end;

--------------------------------------------------------------------------------
-- Strict version of AnimPlay.
-- @param func_orig original function.
-- @param object object name, string.
-- @param animfunc animation function, string.
-- @param func_end animation end handler.
--------------------------------------------------------------------------------
function private.AnimPlay ( func_orig, object, animfunc, func_end )
  
  local who = "AnimPlay";
  
  -- Check 'object'
  if not ( private.IsStr( object ) ) then
    
    private.TraceBadType( who, "object name", object );
    
  elseif not ( private.IsObjExists( object ) ) then
    
    private.TraceError( who, "object does not exist: '" .. object .. "'" );
    
  end;
  
  -- Check 'animfunc'
  if not ( private.IsStr( animfunc ) ) then
    
    private.TraceBadType( who, "animfunc", object );
    
  end;
  
  -- Check 'func_end'
  if not ( private.CheckParam_Handler( func_end ) ) then
    
    private.TraceBadType( who, "func_end", object );
    
  end;
  
  -- Pass data to original API
  return func_orig( object, animfunc, func_end );
  
end;

--------------------------------------------------------------------------------
-- Strict version of DbExecuteQuery.
-- Can trace request type, DB, time, request and result data, total time.
-- @param func_orig original function.
-- @param query query table.
--------------------------------------------------------------------------------
function private.DbExecuteQuery ( func_orig, query )
  
  local who = "DbExecuteQuery";
  local msg = "\n";
  
  private.queries_count = private.queries_count + 1;

  -- Tarce filter
  local trace = true;
  if ( private.trace_queries_filter ) then

    trace = false;

    local type_regex       = private.trace_queries_filter_type;
    local collection_regex = private.trace_queries_filter_collection;

    if ( string.match( query.query,      type_regex       ) ) and
       ( string.match( query.collection, collection_regex ) ) then

      trace = true;

    end;

  end;
  
  if ( trace and private.trace_queries_type ) then
    
    msg = msg .. "Query '" .. query.query      .. "' "
              .. "in DB '" .. query.collection .. "'\n";
    
  end;
  
  local time_before = os.clock();
  local result      = func_orig( query );
  local time_after  = os.clock();
  
  local time = time_after - time_before;
  private.queries_time = private.queries_time + time;

  
  if ( trace ) then

    if ( private.trace_queries_data ) then
    
      msg = msg .. "Query data: \n"
                .. "======================================="
                .. private.AnyToStr( query )
                .. "-> -> -> -> -> -> -> -> -> -> -> -> -> "
                .. private.AnyToStr( result )
                .. "======================================="
                .. "\n";
      
    end;

    if ( private.trace_queries_time ) then
      
      msg = msg .. "Query time: " .. time  .. " sec.\n";
      
    end;

    if ( private.trace_queries_total ) then
      
      msg = msg .. "Req  total: " .. private.queries_count .. "\n"
                .. "Time total: " .. private.queries_time  .. " sec.\n"
      
    end;
    
    private.TraceInfo( who, msg );

  end;
  
  return result;
  
end;

--------------------------------------------------------------------------------
-- function ... () [ SPACER ]
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Strict version of ObjSet.
-- @param func_orig original function.
-- @param object object name, string.
-- @param params object parmaeters, table (see wiki).
--------------------------------------------------------------------------------
function private.ObjSet ( func_orig, object, params )
  
  local who = "ObjSet";
  
  -- Check 'object'
  if not ( private.IsStr( object ) ) then
    
    private.TraceBadType( who, "object name", object );
    
  elseif not ( private.IsObjExists( object ) ) then
    
    private.TraceError( who, "object does not exist: '" .. object .. "'" );
    
  end;
  
  -- Check 'params'
  if not ( private.IsTab( params ) ) then
    
    private.TraceBadType( who, "parameters table", params );
    
  elseif ( private.pedantic ) then
    
    private.ObjSet_CheckParams( object, params, who );
    
  end;
  
  -- Pass data to original API
  return func_orig( object, params );
  
end;

--------------------------------------------------------------------------------
-- Object parameters table checker.
-- Trace errors if invalid parameters founded.
-- @param obj_name object name.
-- @param params parameters table.
-- @param who caller name, e.g. 'ObjSet' or 'ObjMultiset'.
--------------------------------------------------------------------------------
function private.ObjSet_CheckParams ( obj_name, params, who )
  
  private.TraceDepthMore();
  
  if not ( private.IsStr( obj_name ) ) then
    
    obj_name = "obj_unknown";
    
  end;
  
  for param, value in pairs( params ) do
    
    local checker  = private.ObjSet_GetParamChecker( obj_name, param );
    local par_type = "parameter";
    
    -- Check unknown parameter as tag
    if not ( checker ) then
      
      if not ( private.IsTaggedObject( obj_name ) ) then
        
        private.TraceError( who, "try to set tag "
                                 .. "'" .. param .. "'"
                                 .. " for non-animation object: "
                                 .. "'" .. obj_name .. "'" );
        
      end;
      
      checker = private.ObjSet_CheckTag;
      par_type = "tag";
      
    end;
    
    if not ( checker( value ) ) then
      
      private.TraceError( who, "bad " .. par_type .. " value: " .. param 
                               .. " = " .. tostring( value ) );
    end;
  
  end;

  private.TraceDepthLess();
  
end;

--------------------------------------------------------------------------------
-- Parameter checker getter.
-- @param obj_name object name.
-- @param param parameter name.
-- @return parameter checker function.
--------------------------------------------------------------------------------
function private.ObjSet_GetParamChecker ( obj_name, param )

  local obj_type    = sml.obj.GetType( obj_name ) or "__obj_does_not_exists";
  local check_table = private.checkers[ obj_type ];
  
  if not ( check_table ) then
    
    local msg = "[ SAFE API ][ WARNING ] unknown type '%s' of object '%s'";
    DbgTrace( msg:format( obj_type, obj_name ) );
    check_table = private.checkers.obj;
    
  end;
  
  return check_table[param];
  
end;

--------------------------------------------------------------------------------
-- Check if value is valid tag value.
-- @return true if value is string or number.
--------------------------------------------------------------------------------
function private.ObjSet_CheckTag ( value )
  
  return ( private.IsNum( value ) or private.IsStr( value ) );
  
end;

--------------------------------------------------------------------------------
-- function ... () [ SPACER ]
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Strict version of ObjAnimate.
-- @param func_orig original function.
-- @param object object name, string.
-- @param param animated parameter, like 'pos_x'.
-- @param is_loop loop animation flag.
-- @param is_relative relative animation flag (see wiki).
-- @param endtrig after animation played trigger.
-- @param animtable animation table.
--------------------------------------------------------------------------------
function private.ObjAnimate( 
    func_orig, object, param, is_loop, is_relative, endtrig, animtable
  )
  
  local who = "ObjAnimate";
  
  -- Check 'object'
  if not ( private.IsStr( object ) ) then
    
    private.TraceBadType( who, "object name", object );
    
  elseif not ( private.IsObjExists( object ) ) then
    
    private.TraceError( who, "object does not exist: '" .. object .. "'" );
    
  end;
  
  -- Check 'param' and get checker for animtable content checks
  local anim_param_count = 1;
  local param_checker = nil;
  
  if not ( private.IsStr( param ) ) then
    
    private.TraceBadType( who, "animation parameter name", param );

  else

    -- Try to get custom animation param, if non exists try to get non-custom
    param_checker = private.ObjSet_GetParamChecker( object, param );
    
    if not ( param_checker ) then
      
      local non_custom_anim = private.non_custom_anims[param];
      if ( non_custom_anim ) then
        
        anim_param_count = non_custom_anim[1];
        param_checker    = non_custom_anim[2];
        
      end;
      
    end;
    
  end;

  -- Check 'is_loop'
  if not ( private.IsNum( is_loop ) ) then
    
    private.TraceBadType( who, "loop flag", is_loop );
    
  elseif not ( is_loop == 0 or is_loop == 1 ) then
    
    private.TraceError( who, "bad loop flag value: " .. tostring( is_loop ) );
    
  end;
  
  -- Check 'is_relative'
  if not ( private.IsNum( is_relative ) ) then
    
    private.TraceBadType( who, "relative flag", is_relative );
    
  elseif not ( is_loop == 0 or is_loop == 1 ) then
    
    private.TraceError( who, "bad relative flag value: "
                             .. tostring( is_relative ) );
    
  end;

  -- Check 'endtrig'
  if not ( private.CheckParam_Handler( endtrig ) or ( endtrig == nil ) ) then
    
    private.TraceBadType( who, "endtrig", endtrig );
    
  end;
  
  -- Check 'animtable'
  if not ( private.IsTab( animtable ) ) then
    
    private.TraceBadType( who, "animation table", animtable );
  
  elseif ( #animtable < 1 ) then
    
    private.TraceError( who, "empty animation table" );
    
  elseif ( private.pedantic and param_checker ) then
    
    private.ObjAnimate_CheckAnimTable( who, animtable,
                                       param_checker, anim_param_count );
    
  end;
  
  -- Pass data to original API
  return func_orig( object, param, is_loop, is_relative, endtrig, animtable );
  
end;

--------------------------------------------------------------------------------
-- Animation table parameters checker. Trace errors if animatable is invalid.
-- @param who error trace sender, string.
-- @param animtable animation table, table.
-- @param checker value checking function.
-- @param count animation parameters count.
--------------------------------------------------------------------------------
function private.ObjAnimate_CheckAnimTable ( who, animtable, checker, count )
  
  private.TraceDepthMore();
  
  -- animtable entry len
  local entry_len = count + 2;
  
  if ( ( #animtable % entry_len ) ~= 0 ) then
    
    private.TraceError( who, "bad animtable length" );
    
  end;
    
  for frame = 1, #animtable, entry_len  do
    
    local time   = animtable[frame];
    local interp = animtable[frame + 1];
    
    -- Setup values table
    local values = {};
    for i = 1, count do
      
      local value = animtable[frame + 1 + i];
      
      if not ( value ) then
        
        private.TraceError( who, "nil animation parameter value"
                                 .. " at position "
                                 .. tostring( frame + 1 + i ) );
        
      end;
      
      table.insert( values, value );
      
    end;
    
    -- Check 'time'
    if not ( private.IsNum( time ) ) then
      
      private.TraceError( who, "key frame time is not a number: "
                               .. type( time ) );
      
    elseif ( time < 0 ) then
      
      private.TraceError( who, "bad key frame time value: "
                                .. tostring( time ) );
      
    end;
    
    -- Check 'interp'
    if not ( private.IsNum( interp ) ) then
      
      private.TraceError( who, "interpolation type is not a number: "
                               .. type( interp ) );
      
    elseif ( ( interp < 0 ) or ( interp > 3 ) ) then
      
      private.TraceError( who, "bad interpolation type value: "
                               .. tostring( interp ) );
      
    end;

    -- Check all values
    for i = 1, #values do
      
      if not ( checker( values[i] ) ) then
      
        private.TraceError( who, "bad animated parameter value: "
                                 .. tostring( values[i] )
                                 .. " (at position: "
                                 .. tostring( frame + 1 + i )
                                 .. ")" );
      end;
      
    end;
    
  end;

  private.TraceDepthLess();
  
end;

--------------------------------------------------------------------------------
-- function ... () [ SPACER ]
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Strict version of GetObjPosByObj.
-- @param func_orig original function.
-- @param objname object name.
-- @param[opt] baseobj base object name, default value 'ng_application'.
-- @return table { x, y } or empty table, if object does not exists.
--------------------------------------------------------------------------------
function private.ObjPosByObj( func_orig, objname, baseobj )

  local who = "GetObjPosByObj";
  
  -- Check 'objname'
  if not ( private.IsStr( objname ) ) then
    
    private.TraceBadType( who, "object name", objname );
    
  elseif not ( private.IsObjExists( objname ) ) then
    
    private.TraceError( who, "object does not exist: '" .. objname .. "'" );
    
  end;

  -- Check 'baseobj'
  if ( baseobj ) then

    if not ( private.IsStr( baseobj ) ) then
      
      private.TraceBadType( who, "base object name", baseobj );
      
    elseif not ( private.IsObjExists( baseobj ) ) then
      
      private.TraceError( who, "base object does not exist: '" 
                                .. baseobj .. "'" );
      
    end;

  end;

  -- Pass data to original API
  return func_orig( objname, baseobj );

end;

--------------------------------------------------------------------------------
-- function ... () [ SPACER ]
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Strict version of ObjClone.
-- @param func_orig original function.
-- @param objname name of object for clone.
-- @param suffix suffix for new object name.
-- @param clone_childs boolen flag for clone child objects.
-- @param parent parent object name for clone.
--------------------------------------------------------------------------------
function private.ObjClone( func_orig, objname, suffix, clone_childs, parent )

  local who = "ObjClone";

  -- Check 'objname'
  if not ( private.IsStr( objname ) ) then
    
    private.TraceBadType( who, "base object name", objname );
    
  elseif not ( private.IsObjExists( objname ) ) then
    
    private.TraceError( who, "object does not exist: '" .. objname .. "'" );
    
  end;

  -- Check 'suffix'
  if not ( private.IsStr( suffix ) or 
           private.CheckParam_Integer( suffix ) ) then
    
    private.TraceBadType( who, "new object name suffix", suffix );

  end;

  -- Check 'clone_childs'
  if not ( private.CheckParam_Flag( clone_childs ) ) then
    
    private.TraceBadType( who, "clone childs flag", clone_childs );

  end;

  -- Check 'parent'
  if not ( private.IsStr( parent ) ) then
    
    private.TraceBadType( who, "parent object name", parent );
    
  elseif not ( private.IsObjExists( parent ) ) then
    
    private.TraceError( who, "parent does not exist: '" .. parent .. "'" );
    
  end;

  return func_orig( objname, suffix, clone_childs, parent );

end;

--------------------------------------------------------------------------------
-- function ... () [ SPACER ]
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Strict version of ne.mod.Load and ne.mod.LoadLazily.
-- Also set public module table strict.
-- @param func_orig original function.
-- @param path path to module.
--------------------------------------------------------------------------------

function private.ModLoad ( func_orig, path )

  local who = "ModLoad";

  if not ( private.IsStr( path ) ) then
    
    private.TraceBadType( who, "module path", path );
    
  end;
  
  func_orig( path );
  
  local module_name = path:match( "^.*/mod_(.*)$" );
  
  if ( _G[ module_name ] and _G[ module_name ].__Private ) then
    
    _PRIVATE[ module_name ] = _G[ module_name ].__Private();
    -- dbg_strict.Setup( _G[ module_name ].__Private(), "private" );
  
  end;
  
  if ( _G[ module_name ] and _G[ "dbg_strict" ] ) then
    
    dbg_strict.Setup( _G[ module_name ], module_name );
  
  end;

end;

--------------------------------------------------------------------------------
-- function ... () [ SPACER ]
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Convert any variable to string.
-- @param any variable of any type.
-- @return variable value converted to string.
--------------------------------------------------------------------------------
function private.AnyToStr( any )
  
  local result = "";
  
  if ( private.IsStr( any ) ) then
    
    result = any;
    
  elseif ( private.IsTab( any ) ) then
    
    result = private.TabToStr( any );
    
  else
    
    result = tostring( any );
    
  end;
  
  return result;
  
end;

--------------------------------------------------------------------------------
-- Convert table to string (recursive function).
-- @param tab converting table.
-- @param[opt] depth table depth ( need for indent in recursive call ).
-- @param[opt] newline  newline symbol ('\n' default).
-- @param[opt] indent_step indent string  ('\t' default).
-- @return table converted to string.
--------------------------------------------------------------------------------
function private.TabToStr ( tab, depth, newline, indent_step )

  -- Return enmpty string, if empty or not-table passed
  if not ( tab and private.IsTab( tab ) ) then
    return "";
  end;

  -- Set default data 
  depth       = depth       or 0;
  newline     = newline     or "\n";
  indent_step = indent_step or "\t";

  -- Set indent
  local indent = "";
  for d = 1, depth do
    indent = indent .. indent_step;
  end;
  
  -- Set output string
  local output = "";
  if ( depth == 0 ) then
    output = newline;
  end;

  -- Output table fields as strings
  for key, value in pairs( tab ) do

    if ( private.IsStr( value ) ) then
      value = "'" .. value .. "'";
    end;

    -- Add to output field info string
    output = output .. indent .. tostring( key   ) .. " = " 
                              .. tostring( value ) .. newline;

    -- Recursive run for inner tables
    if ( private.IsTab( value ) ) then
      output = output .. private.TabToStr( value,   depth + 1, 
                                           newline, indent_step );
    end;

  end;

  return output;

end;

--------------------------------------------------------------------------------