-- name=dbg_strict
--------------------------------------------------------------------------------
-- Undefined global variables notifier.
-- Has no public interface, just run Init for enable error tracing.
-- Debug module required.
-- @module dbg_strict
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Setup traces about non-declared global variables usage.
--------------------------------------------------------------------------------
function public.__InitModule ()
  
  public.Setup( _G, "_G" );
  
end;

--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Setup traces about non-declared fields usage in specified table.
-- @param tab_pointer pointer to the strict table.
-- @param tab_name name of the strict table (module name).
--------------------------------------------------------------------------------
function public.Setup ( tab_pointer, tab_name )

  -- Set new metatable
  local mt = getmetatable( tab_pointer );
  if ( mt == nil ) then
    mt = {}; setmetatable( tab_pointer, mt );
  end;

  -- Declared global varibales array
  mt.__declared = {};
  
  -- Redefine access to global variables table indexes
  mt.__index = function (t, n)
  
    local declared = mt.__declared[n];
    
    local is_global = ( t == _G );
    local ignore = 
    { 
      parts_app_wasrestarted =     is_global,
      Init                   = not is_global,
      ConnectionLost         = not is_global,
      ConnectionRestored     = not is_global,
      FeatureUpdated         = not is_global
    };
  
    local info = private.GetInfo();
    
    if ( info.what ~= "C" and 
         not declared and 
         not ignore[ n ] and
         not n:match( "v_" ) and
         not n:match( "lib_" ) and
         not n:match( "int_" ) ) then
    
      local msg = "%s.%s is undefined!";
      private.TraceError( info.src, info.line, msg:format( tab_name, n ) );
      
    end;
    
    return rawget( t, n );
    
  end;

  -- Redefine new index creation in global variables table
  mt.__newindex = function (t, n, v)

    if not ( mt.__declared[n] ) then
    
      local info = private.GetInfo();
      local w = info.what;
      
      if ( w ~= "main" and w ~= "C" and v == nil ) then
      
        local msg = "nil-initialization rejected for %s.%s";
      
        private.TraceError( info.src, info.line, msg:format( tab_name, n ) );
        return;
      
      end;
      
      -- Set field declared
      mt.__declared[n] = true;
      
    end;
    
    rawset( t, n, v );

  end;

end;

--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Get info about function with nil-declaration of new global variable.
-- @return table with fields: 'what', 'source', 'currentline'.
--------------------------------------------------------------------------------
function private.GetInfo ()
  
  -- getinfo arguments
  local level = 3;    -- 0 = getinfo(),
                      -- 1 = private.GetInfo(),
                      -- 2 = __index(),
                      -- 3 = function, that contains the variable

  local what  = "Sl"; -- "S" select t += { source, short_src, linedefined, 
                      --                   lastlinedefined, what           }
                      -- "l" select t += { currentline }
                      -- t.what can be: 
                      --   "Lua"  -- a regular Lua function, 
                      --   "C"    -- it is a C function, 
                      --   "main" -- it is the main part of a Lua chunk

  -- NOTE: debug functions so slow
  local info = debug.getinfo( level, what );
  
  if not ( info ) then info = {}; end;
  
  local ret_info = {
    what = info.what        or "C",
    src  = info.source      or "Unknown source",
    line = info.currentline or -1
  };
  
  return ret_info;

end;

--------------------------------------------------------------------------------
-- Trace error with known callpoint (module:string).
-- @param src  source of error (module name).
-- @param line line in source.
-- @param msg  error message.
--------------------------------------------------------------------------------
function private.TraceError ( src, line, msg )
  
  Log.e( src .. ":" .. line .. " " .. msg );
  
end;

--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------
function public.StartArchitect ()
  
  local method_count = 0;
  
  local arch = {};
  
  arch.modules = {
    lib = {},
    m   = {},
    v   = {},
    int = {},
    ui  = {}
  };
  
  local scope = {
    lib = true,
    m   = true,
    v   = true,
    int = true,
    ui  = true
  };

  --- JK: sorting modules
  for global_name, global_data in pairs( _G ) do
    
    if ( scope.lib ) and ( global_name:find( "lib_" ) ) then
      
      table.insert( arch.modules.lib, global_name );
      
    end;
    
    if ( scope.m ) and ( global_name:find( "m_" ) ) then
      
      table.insert( arch.modules.m, global_name );
      
    end;
    
    if ( scope.v ) and ( global_name:find( "v_" ) ) then
      
      table.insert( arch.modules.v, global_name );
      
    end;
    
    if ( scope.int ) and ( global_name:find( "int_" ) ) then
      
      table.insert( arch.modules.int, global_name );
      
    end;
    
  end;

  if ( scope.ui ) then
    
    table.insert( arch.modules.ui, "ui" );
    
  end;
  
  --- JK: injecting
  for _, category in pairs( arch.modules ) do
    
    for _, module_name in pairs( category ) do
    
      local global_data = _G[ module_name ];
      
      --DbgTrace( module_name )
      
      for method_name, method_data in pairs( global_data ) do
      
        if ( type( method_data ) == "function" ) then
          
          local raw_func = method_data;
          method_count   = method_count + 1;
          
          --- PROXY FUNCTION ---------------------------------------------------
          
          _G[ module_name ][ method_name ] = function ( ... )
            
            local args  = { ... };
            local stack = {
              --level1 = debug.getinfo( 1, "Sln" ) or {},
              level2 = debug.getinfo( 2, "Sln" ) or {},
              level3 = debug.getinfo( 3, "Sln" ) or {}
            };
            
            local stack_item = tern( stack.level2.source == "dbg_strict" )(
              stack.level3,
              stack.level2
            );
            
            private.InspectCall(
              module_name,
              method_name,
              args,
              stack_item
            );
            
            local result1, result2, result3, result4, result5 = raw_func( ... );
            
            return result1, result2, result3, result4, result5;
            
          end;
            
          ----------------------------------------------------------------------
          
        elseif ( type( method_data ) == "table" ) then
          
          local method_group = method_data;
          
          for method_subname, method_subdata in pairs( method_group ) do
            
            if ( type( method_subdata ) == "function" ) then
            
              local raw_func = method_subdata;
              method_count   = method_count + 1;
              
              --DbgTrace( method_name.."."..method_subname )
              
              --- PROXY FUNCTION -----------------------------------------------
              
              _G[ module_name ][ method_name ][ method_subname ] = function ( ... )
                
                local args  = { ... };
                local stack = {
                  --level1 = debug.getinfo( 1, "Sln" ) or {},
                  level2 = debug.getinfo( 2, "Sln" ) or {},
                  level3 = debug.getinfo( 3, "Sln" ) or {}
                };
                
                local stack_item = tern( stack.level2.source == "dbg_strict" )(
                  stack.level3,
                  stack.level2
                );
                
                local full_method_name = method_name.."."..method_subname;
                local full_global_name = module_name.."."..method_name.."."..method_subname;
                
                if ( full_global_name == "ui.anim.Play" ) then
                  
                  private.InspectAnimPlay( args, stack_item );
                  
                end;
                
                private.InspectCall(
                  module_name,
                  full_method_name,
                  args,
                  stack_item
                );
                
                local result1, result2, result3, result4, result5 = raw_func( ... );
                
                if ( full_global_name == "common.sound.Play" ) and
                   ( not result1                             ) then
                  
                  local target_sound      = args[ 1 ];
                  local alternative_sound = args[ 4 ];

                  DbgTrace(
                    "[ ERROR ] Failed to play both sounds: "..
                    target_sound..", "..
                    alternative_sound
                  );
                  
                end;
            
                return result1, result2, result3, result4, result5;
                
              end;
                
              ------------------------------------------------------------------
            
            end;
          
          end;
          
        end;
        
      end;
      
    end;
    
  end;
  
  DbgTrace(
    "[ STRICT ] Inspecting "..method_count.." public methods. "..
    "[ Stats ] "..
    "lib: "..#arch.modules.lib..", "..
    "m: "  ..#arch.modules.m..", "..
    "v: "  ..#arch.modules.v..", "..
    "int: "..#arch.modules.int..", "..
    "ui: " ..#arch.modules.ui
  );
  
end;
--------------------------------------------------------------------------------
function private.InspectCall( module_name, method_name, args, stack_item )
  
  -- JK: no info on calling module
  if ( not stack_item.source ) then return; end;
  
  local called_module = {
    name     = module_name,
    method   = method_name,
    is_lib   = module_name:find( "lib_" ),
    is_model = module_name:find( "m_" ),
    is_view  = module_name:find( "v_" ),
    is_int   = module_name:find( "int_" ),
    is_ui    = module_name:find( "ui" )
  };  

  local calling_module = {
    name     = stack_item.source,
    method   = stack_item.name or "?",
    is_lib   = module_name:find( "lib_" ),
    is_model = stack_item.source:find( "m_" ),
    is_view  = stack_item.source:find( "v_" ),
    is_int   = stack_item.source:find( "int_" ),
    is_ui    = ( stack_item.source == "ui" ),
    line     = stack_item.currentline
  };

  ------------------------------------------------------------------------------
  -- IGNORE LIST
  ------------------------------------------------------------------------------

  -- JK: loader is ignored
  if ( calling_module.name == "loader" )           then return; end;

  -- JK: cheater is ignored
  if ( calling_module.name == "cheater" )          then return; end;

  -- JK: calling module method is proxied [?]
  if ( calling_module.method == "raw_func" )       then return; end;

  -- JK: module calls self
  if ( called_module.name == calling_module.name ) then return; end;

  -- JK: mediator is acceptable
  if ( calling_module.name == "mediator" )         then return; end;

  ------------------------------------------------------------------------------

  local func_ThrowWarning = function ()

    DbgTrace(
      "[ WARNING ] "..calling_module.name..":"..calling_module.line.." -> "..
      called_module.name.."."..method_name
    );
    
    --DbgTrace( "[ ARCH ] "..module_name.."."..method_name.." call." );
    --DbgTrace( args );
    --DbgTrace( stack_item );

  end;

  -- JK: inspecting library method call ----------------------------------------
  -- Libraries can only be called from models
  if ( called_module.is_lib ) then

    if ( not calling_module.is_model ) then

      func_ThrowWarning();
      return;

    end;

    return;

  end;

  -- JK: inspecting model method call ------------------------------------------
  -- Models can only be called from their views
  if ( called_module.is_model ) then

    if ( not calling_module.is_view ) then

      func_ThrowWarning();
      return;

    end;

    local model_real_name = called_module.name:gsub( "m_", "" );
    local view_real_name  = calling_module.name:gsub( "v_", "" );

    if ( model_real_name ~= view_real_name ) then

      local dependencies = private.GetDependencies( calling_module.name );

      if ( dependencies[ called_module.name ] ) then

        return;

      end;

      func_ThrowWarning();
      return;

    end;

    return;

  end;

  -- JK: inspecting view method call -------------------------------------------
  -- Views can only be called from other views, common
  if ( called_module.is_view ) then

    if ( not calling_module.is_view          ) and
       ( not calling_module.name == "common" ) then

      func_ThrowWarning();
      return;

    end;

    return;

  end;

  -- JK: inspecting int method call --------------------------------------------
  -- Ints can only be called from relevant views or ui
  -- Exclusion: __InitObjects called from C
  -- Exclusion: __GetDelegateScheme called from common.delegate
  if ( called_module.is_int ) then

    if ( called_module.method == "__InitObjects"       ) or
       ( called_module.method == "__GetDelegateScheme" ) then

      return;

    end;

    if ( not calling_module.is_view ) and
       ( not calling_module.is_ui   ) then

      func_ThrowWarning();
      return;

    end;

    local dependencies = private.GetDependencies( calling_module.name );

    if ( dependencies[ called_module.name ] ) then

      return;

    end;

    func_ThrowWarning();

    return;

  end;

  -- JK: inspecting ui method call ---------------------------------------------
  -- UI can only be called from views, ints and common  
  if ( called_module.is_ui ) then

    if ( not calling_module.is_view          ) and
       ( not calling_module.is_int           ) and
       ( not calling_module.name == "common" ) then

      func_ThrowWarning();
      return;

    end;

    return;

  end;

end;
--------------------------------------------------------------------------------
function private.InspectAnimPlay ( args, stack_item )
  
  local anm_name,
        main_animfunc,
        func_end,
        trailing_animfunc = args[ 1 ], args[ 2 ], args[ 3 ], args[ 4 ];
        
  
  local obj = ne.obj.Get( anm_name );
        
  if ( not obj ) then
    Log.e( anm_name.." doesn't exist" );
    return;
  end;
  
  local animfuncs = obj.anim_funcs;
  
  if ( not common.IsEntityInList( animfuncs, main_animfunc, nil, true ) ) then
    
    local res = ne.obj.Get( anm_name ).res;
    
    local call = {
      name   = stack_item.source,
      method = stack_item.name or "?",
      line   = stack_item.currentline
    };
    
    DbgTrace(
      "[ ERROR ] "..call.name.."."..call.method..":"..call.line.." -> "..
      "failed to play animfunc '"..main_animfunc..
      "' from file: '"..res.."'"
    );
    
  end;
  
  if ( not func_end ) then return; end;
  
  if ( type( func_end ) ~= "function" ) then
  
    local call = {
      name   = stack_item.source,
      method = stack_item.name or "?",
      line   = stack_item.currentline
    };
    
    DbgTrace(
      "[ ERROR ] "..call.name.."."..call.method..":"..call.line.." -> "..
      "endtrig for '"..main_animfunc.."' is not a function!"
    );
    
  end;
  
  if ( not trailing_animfunc ) then return; end;
  
  if ( not common.IsEntityInList( animfuncs, trailing_animfunc, nil, true ) ) then
    
    local call = {
      name   = stack_item.source,
      method = stack_item.name or "?",
      line   = stack_item.currentline
    };
    
    DbgTrace(
      "[ ERROR ] "..call.name.."."..call.method..":"..call.line.." -> "..
      "failed to play animfunc '"..trailing_animfunc..
      "' from file: '"..obj.res.."'"
    );
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.GetDependencies ( module_name )

  if ( _G[ module_name ].__GetDependencies ) then

    local index = _G[ module_name ].__GetDependencies();
    local assoc = {};

    for _, dependency in pairs( index ) do

      assoc[ dependency ] = dependency;

    end;

    return assoc;

  end;

  return {};

end;
--------------------------------------------------------------------------------