-- name=lib_content
--------------------------------------------------------------------------------
-- Background content download module
--------------------------------------------------------------------------------
function public.__InitModule ()

  public.msg =
  {
    DownloadStarted  = "Library_Content_DownloadStarted",
    DownloadProgress = "Library_Content_DownloadProgress",
    DownloadFinished = "Library_Content_DownloadFinished",
    UnpackProgress   = "Library_Content_UnpackProgress"
  };
  
  private.download_total = 0;
  private.download_count = 0;
  private.downloading = false;

  -- Packet download retries
  private.retry_packet = "";
  private.retry_number = 0;
  private.retry_timer  =
  {
    token = "tmr_content_download_retry",
    scope = "SESSION_GLOBAL_TIME",
    time  = 5,
    event = "Library_Content_DownloadRetry"
  };

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Silent_IsDownloading ()

  return private.downloading;

end;
--------------------------------------------------------------------------------
function public.Silent_IsReady ()

  if ( private.IsSilentPacketsExists() ) then
  
    local list = private.GetPacketList();
    return ( #list < 1 );
  
  end;

  return false;

end;
--------------------------------------------------------------------------------
function public.Silent_HasFreeSpace ()
  
  local total   = 0;
  local packets = parts.content.GetPacketList();
  
  for _, packet in pairs( packets ) do
  
    local packet_number = tonumber( packet.name:match( "_(%d+)" ) );

    if ( not packet.force_update and
         packet.size and
         ( not packet_number or packet_number <= ( private.sections_allowed or 1 ) ) and
         packet.state == "downloading" ) then

      total = total + ( math.tointeger( packet.size ) or 0 );
    
    end;
  
  end;  
  
  local total_mb = math.ceil( total / 1000000 );

  return ne.file.HasFreeSpace( "save/", total ), total_mb;
  
end;
--------------------------------------------------------------------------------
function public.Silent_StartDownloading ()

  private.sections_allowed = m_game.Section_GetAllowedCount();
  
  local list = private.GetPacketList();
  if ( #list > 0 ) then
  
    private.download_total = #list;
    private.download_count = 0;
    
    private.TryDownloadNext();
  
  end;

end;
--------------------------------------------------------------------------------
function public.Silent_ApplyPackets ()

  private.Silent_ApplyPackets();

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Packet_IsSilent ( packet_name, packet_data )

  return private.Packet_IsSilent( packet_name, packet_data );

end;
--------------------------------------------------------------------------------
function public.Packet_IsMinievent ( packet_name )

  return private.Packet_IsMinievent( packet_name );

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.Packet_IsSilent ( packet_name, packet_data )
  
  if ( not packet_data ) then
    
    local packets = parts.content.GetPacketList();
    
    for _, packet in pairs( packets ) do
      
      if ( packet.name == packet_name ) then
      
        packet_data = packet;
        
        break;
      
      end;
    
    end;

  end;
  
  if ( not packet_data.force_update ) then
    
    return true;
    
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
function private.Packet_IsMinievent ( packet_name )

  return packet_name:match( "^(minievent_).*$" ) ~= nil;

end;
--------------------------------------------------------------------------------
function private.IsSilentPacketsExists ()

  local packets = parts.content.GetPacketList();
  
  for _, packet in pairs( packets ) do
    
    if ( private.Packet_IsSilent( packet.name, packet ) ) then
    
      return true;
    
    end;
  
  end;
  
  return false;

end;
--------------------------------------------------------------------------------
-- function ... () end;
--------------------------------------------------------------------------------
function private.GetPacketList ()

  local list    = {};
  local packets = parts.content.GetPacketList();
  
  local packages_priority = {
    audio       = 7,
    boosters    = 4,
    characters  = 4,
    database    = 2,
    enemies     = 4,
    hero        = 4,
    interface   = 3,
    items       = 4,
    levels      = 2,
    loading     = 4,
    map         = 3,
    match3      = 3,
    minievent   = 6,
    scripts     = 0,
    section     = 3,
    skins       = 4,
    stickers    = 5,
    strings     = 1,
    worlds      = 5,
    zones       = 3
  };
  
  for _, packet in pairs( packets ) do
  
    packet.priority = packages_priority[ packet.name ] or
                      packages_priority[ packet.name:match( "(%w+)_%d+" ) ] or
                      6;
  
  end;
                    
  table.sort( packets, function ( a ,b )
  
    return a.priority < b.priority;
    
  end );

  for _, packet in ipairs( packets ) do
  
    local packet_number = tonumber( packet.name:match( "_(%d+)" ) );

    if (   not packet.force_update and
         ( not packet_number or packet_number <= ( private.sections_allowed or 1 ) ) and
         packet.state == "downloading" ) then

      table.insert( list, packet.name );
      loader.Content_SetPacketState( packet.name, "downloading" );
    end;
  
  end;
  
  return list;

end;
--------------------------------------------------------------------------------
function private.Silent_ApplyPackets ( only_minievent )
  
  local packets = parts.content.GetPacketList();

  local func_ApplyPacket = function ( packet_name )
    parts.content.ApplyPacket( packet_name );
    loader.Content_SetPacketState( packet_name, "ready" );
  end;

  for _, packet in pairs( packets ) do
    
    if ( packet.name                  ) and
       ( packet.state                 ) and
       ( packet.state == "downloaded" ) and 
       ( packet.downloaded            ) then

      if ( only_minievent ) then
        if ( private.Packet_IsMinievent( packet.name ) ) then
          func_ApplyPacket( packet.name );
        end;
      else
        func_ApplyPacket( packet.name );
      end;

    end;
    
  end;
  
end;
--------------------------------------------------------------------------------
function private.TryDownloadNext ()

  local list = private.GetPacketList();
  local packet = list[ 1 ];
  
  if ( packet ) then
    
    private.downloading = true;
    mediator.Notify( public.msg.DownloadStarted );
    
    wiz.StartBackground();
    ContentStartDownloading( packet, private.DownloadCallback );
  
  end;

end;
--------------------------------------------------------------------------------
function private.DownloadCallback ( name, status, total, progress )
  
  if     ( status == "progress"  ) then
    
    local notify_data =
    {
      current  = private.download_count + 1,
      total    = private.download_total,
      progress = math.floor( progress / total * 100 ),
      name = name
    };
    mediator.Notify( public.msg.DownloadProgress, notify_data );

  elseif ( status == "unpacking" ) then
  
    local notify_data =
    {
      current  = private.download_count + 1,
      total    = private.download_total,
      progress = math.floor( progress / total * 100 ),
      name = name
    };
    mediator.Notify( public.msg.UnpackProgress, notify_data );

  elseif ( status == "success"   ) then
  
    private.download_count = private.download_count + 1;
  
    if ( private.download_total == private.download_count ) then
    
      wiz.EndBackground();
      private.downloading = false;
      
      private.Silent_ApplyPackets( true );
      mediator.Notify( public.msg.DownloadFinished, true );
      
    else
    
      private.TryDownloadNext();
    
    end;

  elseif ( status == "failed"    ) then

    if ( private.retry_packet ~= name ) then

      private.retry_packet = name;
      private.retry_number = 0;

    end;

    if ( private.retry_number > 2 ) then

      wiz.EndBackground();
      private.Silent_ApplyPackets( true );
      private.downloading = false;
      private.retry_packet = "";
      private.retry_number = 0;
      mediator.Notify( public.msg.DownloadFinished, false );
      return;

    end;

    private.retry_number = private.retry_number + 1;

    local function func_retry ()
      parts.time.event.Finish();
      mediator.Unsubscribe( private.retry_timer.event, func_retry );
      private.TryDownloadNext();
    end;
    mediator.Subscribe( private.retry_timer.event, func_retry );
    
    parts.time.event.Shedule(
      private.retry_timer.token, 
      private.retry_timer.scope,
      private.retry_timer.time,
      private.retry_timer.event,
      {}
    );

  end;
  
end;
--------------------------------------------------------------------------------