-- name=lib_rpc
--------------------------------------------------------------------------------
-- Remote procedure call library.
-- @module lib_rpc
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Module constructor.
--------------------------------------------------------------------------------
function public.__InitModule ()

  private.scheduled = {};
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Setup ( callbacks )
  
  private.WaitingShow = callbacks.WaitingShow; -- Show waiting splash
  private.WaitingHide = callbacks.WaitingHide; -- Hide waiting splash
  
end;

--------------------------------------------------------------------------------
-- Schedule remote procedure call with the next HeartBeat.
-- @param target server handler name
-- @param method server handler method name
-- @param params server handler method parameters
-- @param func_handle handler for the returned data
-- @param is_silent do not show waiting splash until remote procedure executed
-- @param restart_heartbeat restart HeartBeat immediately
--------------------------------------------------------------------------------
function public.Schedule (
    target,
    method,
    params,
    func_handle,
    is_silent,
    restart_heartbeat
  )

  local message = private.GetMessage_OnResult( target, method );
  
  if ( not private.scheduled[ message ] ) then
  
    private.scheduled[ message ] =
    {
      target      = target,
      method      = method,
      params      = params,
      func_handle = func_handle,
      is_silent   = is_silent
    };
    ne.msg.Subscribe( message, private.OnResult );
  
  end;
  
  if ( not is_silent ) then private.WaitingShow(); end;
  
  private.Schedule( target, method, params, restart_heartbeat );

end;

--------------------------------------------------------------------------------
-- Schedule remote procedure call with the next HeartBeat.
-- @param target server handler name
-- @param method server handler method name
-- @param params server handler method parameters
-- @param restart_heartbeat restart HeartBeat immediately
--------------------------------------------------------------------------------
function private.Schedule ( target, method, params, restart_heartbeat )

  if ( restart_heartbeat ) then restart_heartbeat = 1;
  else                          restart_heartbeat = 0;
  end;

  NetInvokeServerMethod(
    private.Decapitalize( target ),
    private.Decapitalize( method ),
    params,
    restart_heartbeat
  );

end;

--------------------------------------------------------------------------------
-- Remote procedure result handler.
-- @param message notification string
-- @param params notification parameters
--------------------------------------------------------------------------------
function private.OnResult ( message, params )

  ne.msg.Unsubscribe( message, private.OnResult );
  
  local call = private.scheduled[ message ];
  private.scheduled[ message ] = nil;
  
  if ( call.func_handle ) then
  
    call.func_handle( message, params );
  
  end;
  
  if ( not call.is_silent ) then private.WaitingHide(); end;

end;

--------------------------------------------------------------------------------
-- Reschedule all remote procedure calls, that not executed on last HeartBeat.
--------------------------------------------------------------------------------
function public.RescheduleAll ()

  for _, call in pairs( private.scheduled ) do
  
    private.Schedule( call.target, call.method, call.params );
  
  end;

end;

--------------------------------------------------------------------------------
-- Remove all scheduled remote procedure calls.
--------------------------------------------------------------------------------
function public.ClearScheduled ()

  for message in pairs( private.scheduled ) do
    ne.msg.Unsubscribe( message, private.OnResult );
  end;
  
  private.scheduled = {};

end;

--------------------------------------------------------------------------------
-- Check if any of scheduled call is not silent.
-- @return true if all calls are silent
--------------------------------------------------------------------------------
function public.IsSilent ()

  for _, call in pairs( private.scheduled ) do
  
    if ( not call.is_silent ) then return false; end;
  
  end;
  
  return true;

end;

--------------------------------------------------------------------------------
-- Create 'on-call-result' message string.
-- @param target server handler name
-- @param method server handler method name
-- @return message that sent at call result received
--------------------------------------------------------------------------------
function private.GetMessage_OnResult ( target, method )

  return "Server_"..target.."_on"..method;

end;

--------------------------------------------------------------------------------
-- Transform first char of call to lower case.
-- @param call_string call
-- @return transformed call string
--------------------------------------------------------------------------------
function private.Decapitalize ( call_string )

  local first = string.sub( call_string, 1, 1 );
  local other = string.sub( call_string, 2    );
  
  return string.lower( first )..other;

end;

--------------------------------------------------------------------------------