-- name=v_characters
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_characters",
  "int_dialog_map_quest_reward"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------

--  Character model states:
--    + locked   -- character not available
--    + unlocked -- character available, but has no quests started
--    + active   -- character available and has active quest
--    + inactive -- character available and all quests done
--
--  Character visual states:
--    + normal -- character in normal state
--    + cage   -- character in caged in cell
--    + charm  -- charcter charmed to ghost
--
--    * 'cage' and 'charm' visual states has same named animation resources,
--      but character behavior and animfuncs are different

--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );

  -- Stateless done
  private.show_states =
  {
    quest  = "qst_collect_gold_key",
    stages = {}
  };

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Init ()

  for s = 1, v_stage.Stateless_Limit() - 1 do

    table.insert( private.show_states.stages, "stage_"..s );

  end;
  
  mediator.Subscribe(
    "Model_Builds_BuildConstructed",
    private.OnBuildConstructed
  );

  mediator.Subscribe(
    "Model_Quests_QuestStarted",
    private.OnQuestStarted
  );

  mediator.Subscribe(
    "Model_Quests_QuestCompleted",
    private.OnQuestCompleted
  );

end;
--------------------------------------------------------------------------------
function private.OnBuildConstructed ( build_data )
  
  -- Try unlock character
  local character = m_characters.GetCharacterByBuild( build_data.name );
  if  ( character ) then

    public.Character_Unlock( character.name );

  end;

end;
--------------------------------------------------------------------------------
function private.OnQuestStarted ( quest_data )

  -- Try activate character
  local character = m_characters.GetCharacterByQuest( quest_data.name );
  if  ( character ) then

    m_characters.SetCharacterState( character.name, "active" );

  end;

end;
--------------------------------------------------------------------------------
function private.OnQuestCompleted ( quest_data )

  -- Try deactivate character
  local character = m_characters.GetCharacterByQuest( quest_data.name );
  if  ( character ) then
  
    local deactivate = false;
    local quest_cage = character.quest_cage;
    local quest_free = character.quest_free;
    
    if ( quest_cage and quest_free ) then
    
      local quest_cage_done = v_quests.Quest_IsComplete( quest_cage );
      local quest_free_done = v_quests.Quest_IsComplete( quest_free );
      
      deactivate = quest_cage_done and quest_free_done;
      
    elseif ( quest_cage ) then
    
      deactivate = v_quests.Quest_IsComplete( quest_cage );
    
    elseif ( quest_free ) then
    
      deactivate = v_quests.Quest_IsComplete( quest_free );
    
    elseif ( IsEditor() ) then
    
      local msg = "Error! Can't check deactivation of %q by quest %q!";
      DbgTrace( msg:format( character.name, quest_data.name ) );
    
    end;
    
    if ( deactivate ) then
    
      m_characters.SetCharacterState( character.name, "inactive" );
    
    end;
    
    v_log.Kontagent_LogEvent( "quest", 
    {
      section = quest_data.section,
      name    = quest_data.name
    } );
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.GetAllInSection ( section_name )

  local  characters = m_characters.GetAllInSection( section_name );
  return characters;

end;
--------------------------------------------------------------------------------
function public.Create ( character, section )

  local obj_name = ( "obj_map_characters_%s" ):format( character.name );
  local btn_name = ( "btn_map_characters_%s" ):format( character.name );
  
  -- ObjCreate( obj_name, "obj"                    );
  ui.obj.Set( obj_name, { anim_tag = character.name } );
  ui.obj.Attach( obj_name, "anm_map_"..section );
  
  ui.obj.Clone( "btn_map_characters", obj_name, "_"..character.name, 1 );
  
  local params = ui.obj.Get( obj_name );
  
  ui.obj.Set( obj_name, { inputrect_init = false } );
  ui.obj.Set( btn_name,
    {
      inputrect_init = true,
      inputrect_x = params.inputrect_x,
      inputrect_y = params.inputrect_y,
      inputrect_w = params.inputrect_w,
      inputrect_h = params.inputrect_h
    }
  );
  
  local func_click = function () public.Tap( character.name ) end;
  ui.control.EventsSet( btn_name, { func_on_click = func_click } );

  -- JK: the character may have chest placed on map beside him
  local has_chest = private.Character_HasChest( character );
  if ( has_chest ) then

    local chest_name = character.chest;
    local anm_name   = "anm_map_chest_"..character.name;

    ui.obj.Clone(
      "anm_map_chest",
      "anm_map_"..section,
      "_"..character.name
    );

    ui.obj.Set(
      anm_name,
      {
        anim_tag = "at_"..character.name.."_chest",
        animfunc = "static_"..chest_name,
        rt_chest = ui.obj.Res( "item", chest_name )
      }
    );

    ui.control.EventsSet(
      anm_name,
      {
        func_on_click = function ()

          ui.input.Off( anm_name );

          ui.anim.Play(
            anm_name,
            "click_"..chest_name,
            function () ui.input.On( anm_name ); end,
            "static_"..chest_name
          );

        end
      }
    );

  end;
  
  local state = m_characters.GetCharacterState( character.name );

  if     ( state == "locked"   ) then private.SetLocked(   character );
  elseif ( state == "unlocked" ) then private.SetUnlocked( character );
  elseif ( state == "active"   ) then private.SetActive(   character );
  elseif ( state == "inactive" ) then private.SetInactive( character );
  end;

end;
--------------------------------------------------------------------------------
function private.SetLocked ( character )

  ui.obj.Disable( "btn_map_characters_"..character.name );

  if ( ui.obj.Exists( "anm_map_chest_"..character.name ) ) then

    ui.obj.Disable( "anm_map_chest_"..character.name );

  end;

end;
--------------------------------------------------------------------------------
function private.SetUnlocked ( character )

  ui.obj.Enable( "btn_map_characters_"..character.name );
  private.Indicator_Disable( character.name, false );

  local state = private.GetVisualState_Unlocked( character.name );
  private.SetVisualState( character.name, state );

  if ( ui.obj.Exists( "anm_map_chest_"..character.name ) ) then

    ui.obj.Enable( "anm_map_chest_"..character.name );

  end;
  
end;
--------------------------------------------------------------------------------
function private.SetActive ( character )
  
  ui.obj.Enable( "btn_map_characters_"..character.name );
  private.Indicator_Enable( character.name, true );
  
  local quest_name = m_characters.GetCurrentQuestName( character.name );
  
  if ( quest_name ) then
    local state = private.GetVisualState_Active( character.name, quest_name );
    private.SetVisualState( character.name, state );
  else
    DbgTrace( "[ ERROR ] SetActive: no quest for character < "..character.name.." >!" );
    private.SetVisualState( character.name, "normal" );
  end;

  if ( ui.obj.Exists( "anm_map_chest_"..character.name ) ) then

    ui.obj.Enable( "anm_map_chest_"..character.name );

  end;
  
end;
--------------------------------------------------------------------------------
function private.SetInactive ( character )
  
  ui.obj.Enable( "btn_map_characters_"..character.name );
  private.Indicator_Disable( character.name );
  
  -- Hack for Kai state.
  if ( character == "kai"          ) or
     ( character == "tombprincess" ) then
    
    private.SetVisualState( character.name, "complete" );
    return;
    
  end;
  
  private.SetVisualState( character.name, "normal" );

  if ( ui.obj.Exists( "anm_map_chest_"..character.name ) ) then

    ui.obj.Enable( "anm_map_chest_"..character.name );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Tap ( character )

  local m_state = m_characters.GetCharacterState( character );
  local v_state = "normal";
  
  if     ( m_state == "unlocked" ) then
  
    v_state = private.GetVisualState_Unlocked( character );
    private.Popup_Unlocked( character );
  
  elseif ( m_state == "active"   ) then
  
    local quest = nil;
    local quest_name = m_characters.GetCurrentQuestName( character );
    if ( quest_name ) then
      quest = v_quests.Quest_Data( quest_name );
    end;
    
    if ( quest ) then
      
      v_state = private.GetVisualState_Active( character, quest.name );
      
      private.Popup_Active(
        character,
        quest,
        function ()
      
          v_map.Point_SuggestByReward( "quest", quest.target.name );
          
        end
      );
      
    else
      
      DbgTrace( "[ ERROR ] Tap: no quest for character < "..character.." >!" );
        
    end;
  
  elseif ( m_state == "inactive" ) then
  
    v_state = tern( ( character == "kai" ) or ( character == "tombprincess" ) )( "complete", v_state );
    private.Popup_Inactive( character );
  
  elseif ( IsEditor()          ) then
  
    local msg = "Error! Character %q in wrong state %q";
    common.DbgCallpoint( msg:format( character, m_state ) );
  
  end;
  
  private.Animate_Action( character, v_state, nil, true );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Character_Unlock ( character_name )

  local character_state = m_characters.GetCharacterState( character_name );

  if ( character_state == "locked" ) then

    m_characters.SetCharacterState( character_name, "unlocked" );

  end;

end;
--------------------------------------------------------------------------------
function public.Character_GetData ( character_name )
  
  local  character_data = m_characters.GetCharacterData( character_name );
  return character_data;
  
end;
--------------------------------------------------------------------------------
function private.Character_HasChest ( character )
  
  if ( not character.chest ) then

    return false;

  end;

  local is_quest_completed = v_quests.Quest_IsComplete( character.quest_free );

  return ( not is_quest_completed );
  
end;
--------------------------------------------------------------------------------
function public.GetCharacterByFragment ( fragment_name )

  local quest = v_quests.Quest_GetNameByItem( "fragment", fragment_name );

  if ( quest ) then

    return m_characters.GetCharacterByQuest( quest );

  end;

  return nil;

end;
--------------------------------------------------------------------------------
function public.GetCharacterByBuild ( build_name )

  local  character = m_characters.GetCharacterByBuild( build_name );
  return character;

end;
--------------------------------------------------------------------------------
function public.IsCharactersQuest ( quest_name )

  return m_characters.GetCharacterByQuest( quest_name ) ~= nil;

end;
--------------------------------------------------------------------------------
function public.GetStageUnlocker( build_name )

  local character = m_characters.GetCharacterByBuild( build_name );

  if ( character ) then

    local quest_name = character.quest_cage or character.quest_free;
    local quest      = v_quests.Quest_Data( quest_name );

    if ( quest ) then

      return tonumber( string.match( quest.condition.name, "%d+" ) );

    end;

  end;

  return false;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Indicator_Enable ( character, show )

  local ind_name = "anm_map_characters_quest_"..character;
  ui.obj.Enable( ind_name, true  );
  
  if ( show ) then
  
    private.Indicator_Show( character );
  
  end

end;
--------------------------------------------------------------------------------
function private.Indicator_Disable ( character_name, hide )

  local ind_name = "anm_map_characters_quest_"..character_name;
  
  if ( hide ) then
  
    private.Indicator_Hide( character_name, function ()
    
      ui.obj.Disable( ind_name, true );
    
    end );
  
  else
  
    ui.obj.Disable( ind_name, true );
  
  end;

end;
--------------------------------------------------------------------------------
function private.Indicator_Show ( character )

  local ind_name = "anm_map_characters_quest_"..character;
  ui.anim.Play( ind_name, "quest_popup_open" );

end;
--------------------------------------------------------------------------------
function private.Indicator_Hide ( character, func_after )

  local ind_name = "anm_map_characters_quest_"..character;
  ui.anim.Play( ind_name, "quest_popup_close", func_after ); 

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Sound_Play ( sound_id )
  
   common.sound.Play( "assets/content/map/sounds/aud_"..sound_id );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
--[[
  Character reaction format:
  {
    -- Basic data:
    name       = "character_name"
    func_popup = function ( func_after ) end,
    func_anim  = function ( func_after ) end,
    
    -- Only for reward drop:
    object          = "obj_char_name", -- receiver-object
    func_drop       = function () end, -- on-drop function
    call_drop_first = true, -- true : func_drop on focus, func_anim on drop
                            -- false: func_anim on focus, func_drop on drop
    
    -- Rewards:
    quest_done    = "completed_quest_name",
    reward        = { ... }, -- quest rewards table
    func_reward   = function ( reactions, func_after ) end,
    skinpart      = "skn_part_name",
    
    -- Other:
    update_stages = { "stage_1", "stage_2", "stage_3", ... }
  }
]]
--------------------------------------------------------------------------------
function public.GetRewardReaction ( stage, rewards )

  local reaction = nil;
  
  if not ( rewards ) then return reaction; end;
  
  -- Find quest item
  local quest_item = nil;
  for i, reward in pairs( rewards ) do
  
    if ( reward.type == "quest" ) then
    
      quest_item = reward;
      break;
    
    end;
  
  end;
  
  if not ( quest_item ) then return reaction; end;
  
  local item_name = quest_item.name;
  
  -- Find character and quest
  local character =
    m_characters.GetCharacterByItemFromStage( item_name, stage );
  
  local quest =
    m_characters.GetQuestByItemFromStage( item_name, stage );
  
  local quest_done = 
    m_characters.IsQuestForItemFromStageDone( item_name, stage );
  
  -- Setup reaction
  reaction        = {};
  reaction.name   = character;
  reaction.object = "obj_map_characters_"..character;
  
  reaction.call_drop_first = false;
  reaction.func_drop       = nil;
  
  reaction.func_popup = function ( func_after )
    private.Popup_ItemReceive( character, quest, func_after );
  end;
  
  if ( quest_done ) then
  
    reaction.quest_done = quest.name;
    reaction.reward = quest.reward;
    
    for i, reward in pairs( quest.reward ) do
    
      if ( reward.type == "skinpart" ) then
      
        reaction.skinpart = reward.name;
        break;
      
      end;
    
    end;
    
    local animation    = private.GetAnimData_QuestDone( character, quest );
    local state        = private.GetVisualState_Complete( character, quest.name );
    reaction.func_anim = v_map.Callback_Animate( animation );
    
    reaction.func_reward = function ( reactions, func_after )
      private.ShowReward( character, state, quest.reward, reactions, func_after );
    end;
    
  else
  
    local animation    = private.GetAnimData_ItemReceive( character, quest );
    reaction.func_anim = v_map.Callback_Animate( animation );
  
  end;
  
  return reaction;

end;
--------------------------------------------------------------------------------
function public.GetQuestCompleteReaction ( quest_name )

  local reaction = nil;
  if not ( quest_name ) then return reaction; end;

  local quest_next = nil;
  local character = m_characters.GetCharacterByQuest( quest_name );
  
  if ( character ) then
  
    if ( quest_name == character.quest_cage ) then
      quest_next = character.quest_free;
    elseif ( quest_name == character.quest_free ) then
      quest_next = character.quest_cage;
    end;
  
  end;
  
  local stateless_condition = ( quest_name == private.show_states.quest );
  
  if ( stateless_condition or 
     ( quest_next and v_quests.Quest_IsCurrent( quest_next ) ) ) then
     
    -- Setup reaction
    reaction        = {};
    reaction.name   = character.name;
    reaction.object = "obj_map_characters_"..character.name;
  
    if ( quest_next ) then
      
      local quest = v_quests.Quest_Data( quest_next );
      
      reaction.func_popup = function ( func_after )
      
        private.Popup_Active( character.name, quest, func_after );
        
      end;
      
      local animation = private.GetAnimData_QuestStart( character.name, quest );
      reaction.func_anim = v_map.Callback_Animate( animation );

      reaction.update_stages = quest.target.list;
      
    end;
    
    if ( stateless_condition ) then
      reaction.update_stages = private.show_states.stages;
    end;
  
  end;
  
  return reaction;

end;
--------------------------------------------------------------------------------
function public.GetStageCompleteReaction ( stage )

  local reaction = nil;
  local character, quest;
  
  local quests = v_quests.GetQuestsOnStageComplete( stage );
  for _, stage_quest in pairs( quests ) do
  
    character = m_characters.GetCharacterByQuest( stage_quest.name );
    if ( character ) then
    
      quest = stage_quest;
      break;
    
    end;
  
  end;
  
  if ( character ) then
  
    local name = character.name;
  
    reaction               = {};
    reaction.name          = name;
    reaction.object        = "obj_map_characters_"..name;
    reaction.update_stages = quest.target.list;
    reaction.func_popup    = function ( func_after )
    
      private.Popup_Active( name, quest, func_after );
      
    end;
    
    local animation    = private.GetAnimData_QuestStart( name, quest );
    reaction.func_anim = v_map.Callback_Animate( animation );
    
  end;
  
  return reaction;

end;
--------------------------------------------------------------------------------
function private.ShowReward (
    character,
    state,
    rewards,
    reactions,
    func_after
  )

  if ( type( rewards ) ~= "table" ) or
     ( #rewards < 1               ) then

    DbgTrace( "[ ERROR ] This character has empty reward: "..character );
    return;

  end;

  if ( rewards[ 1 ].type == "chest" ) then

    local chest_data = rewards[ 1 ];

    private.ShowReward_Chest (
      character,
      state,
      chest_data,
      reactions,
      func_after
    );

  else

    private.ShowReward_Default (
      character,
      state,
      rewards,
      reactions,
      func_after
    );

  end;

end;
--------------------------------------------------------------------------------
function private.ShowReward_Default (
    character,
    state,
    rewards,
    reactions,
    func_after
  )

  local func_close = function ()

    int_dialog_map_quest_reward.Hide(
      function () 
        v_map.Reward_Show( false, rewards, reactions, func_after );
      end, 
      function ()
        v_interface.window.Hide( "int_dialog_map_quest_reward" );
      end
    );

  end;

  local func_start = function ()

    ui.screen.Lock( false );
    v_interface.Hide();

    local character_data = {
      name     = character,
      content  = m_characters.GetContent( character ),
      animfunc = tern( state ~= "complete" )( "static", state )
    };

    int_dialog_map_quest_reward.Show(
      character_data,
      rewards
    );

  end;

  v_interface.window.Show(
    "int_dialog_map_quest_reward",
    func_start, 
    { map_active = true },
    { Accept = func_close }
  );

end;
--------------------------------------------------------------------------------
function private.ShowReward_Chest (
    character,
    state,
    chest_data,
    reactions,
    func_after
  )

  ui.screen.Lock( true );

  local anm_chest   = "anm_map_chest_"..character;
  local ext_rewards = chest_data.items;
  chest_data.items  = nil;

  local obtain_params =
  {
    category   = "character_reward",
    source     = "character_"..character,
    stage_name = "character_"..character,
    silent     = true
  };

  ui.anim.Play(
    anm_chest,
    "get_"..chest_data.name,
    function ()

      v_items.Chest_Show( 
        chest_data,
        obtain_params,
        ext_rewards,
        func_after,
        function () ui.screen.Lock( false ); end
      );

      ui.obj.Disable( anm_chest );

    end
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetMainEnemy ( character_name )

  local section_number = m_characters.GetCharacterData( character_name ).section:match( "%d+" );
  local stage_name     = "stage_"..math.ceil( 110 * section_number );
  local stage_data     = v_stage.Stage_StaticData( stage_name );
  local enemy_name     = stage_data.state[ 1 ].enemy.name;

  return enemy_name.."_"..stage_name;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Popup_Unlocked ( character, func_after )
  
  v_map.ShowObjectPopup(
    "monolog",
    "obj_map_characters_indicator_"..character,
    "str_int_popup_cloud_characters_"..character,
    nil,
    nil,
    {
      func_preclose = func_after
    }
  );
  
end;
--------------------------------------------------------------------------------
function private.Popup_ItemReceive ( character, quest, func_after )

  private.Indicator_Hide( character );

  local func_close = function ()
    private.Indicator_Show( character );
  end;
  
  local data = m_characters.GetQuestPopupData( quest );

  v_map.ShowObjectPopup(
    "counter",
    "obj_map_characters_indicator_"..character,
    "",
    data,
    nil,
    {
      func_preclose = func_after,
      func_close    = func_close
    }
  );

end;
--------------------------------------------------------------------------------
function private.Popup_Active ( character, quest, func_after )

  private.Indicator_Hide( character );
  
  local func_close = function ()

    private.Indicator_Show( character );
    
  end;
  
  local data = m_characters.GetQuestPopupData( quest );

  if ( data.reward ) then
    
    for i = 1, #data.reward do
      
      if ( data.reward[ i ].name == "immortality" ) then
        
        data.reward = data.reward[ i ];

        v_map.ShowObjectPopup(
          "reward",
          "obj_map_characters_indicator_"..character,
          "str_int_popup_cloud_characters_"..character,
          data,
          nil,
          {
            func_preclose = func_after,
            func_close    = func_close
          }
        );
        
        return;
        
      end;
      
    end;
    
  end;
  
  v_map.ShowObjectPopup(
    "complex",
    "obj_map_characters_indicator_"..character,
    "str_int_popup_cloud_characters_"..character,
    data,
    nil,
    {
      func_preclose = func_after,
      func_close    = func_close
    }
  );
  
end;
--------------------------------------------------------------------------------
function private.Popup_Inactive ( character, func_after )
  
  local popup = "str_int_popup_cloud_characters_"..character;
  
  v_map.ShowObjectPopup(
    "monolog",
    "obj_map_characters_indicator_"..character,
    "str_int_popup_cloud_characters_completed_"..character,
    nil,
    nil,
    {
      func_preclose = func_after
    }
  );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.GetVisualState_Unlocked ( character )

  local data       = m_characters.GetCharacterData( character );
  local state_type = data.unlocked_type;
  
  local states =
  {
    cage = "cage",
    free = "normal"
  };
  
  return states[ state_type ];

end;
--------------------------------------------------------------------------------
function private.GetVisualState_Active ( character, quest_name )

  local data = m_characters.GetCharacterData( character );
  
  local quest_cage = data.quest_cage;
  local quest_free = data.quest_free;
  local state_type = data.unlocked_type;
  
  -- DG: Characters data format workaround
  -- Two quests + unlock type free = reverted character states
  if ( quest_cage and quest_free and state_type == "free" ) then
  
    local states =
    {
      [ quest_cage ] = "normal",
      [ quest_free ] = "charm"
    }
    
    return states[ quest_name ];
  
  end;
  
  local states = {};
  
  if ( quest_cage ) then states[ quest_cage ] = "cage";   end;
  if ( quest_free ) then states[ quest_free ] = "normal"; end;
  
  return states[ quest_name ];

end;
--------------------------------------------------------------------------------
function private.GetVisualState_Complete ( character, quest_name )

  local data = m_characters.GetCharacterData( character );
  
  local quest_cage = data.quest_cage;
  local quest_free = data.quest_free;
  local state_type = data.unlocked_type;
  
  -- DG: Characters data format workaround
  -- Two quests + unlock type free = reverted character states
  if ( quest_cage and quest_free and state_type == "free" ) then
  
    local states =
    {
      [ quest_cage ] = "normal",
      [ quest_free ] = "charm"
    }
    
    return states[ quest_name ];
  
  end;
  
  local states = {};
  
  if ( quest_cage ) then states[ quest_cage ] = "cage";   end;
  if ( quest_free ) then states[ quest_free ] = "normal"; end;
  
  if ( ( character == "kai" ) or ( character == "tombprincess" ) ) then states[ quest_free ] = "complete"; end;
  
  return states[ quest_name ];

end;
--------------------------------------------------------------------------------
function private.SetVisualState ( character, state )

  local anm_name = "anm_map_characters_icon_"..character;
  local content = m_characters.GetContent( character );

  local normal =
  {
    res = "assets/content/"..content.."/anm_character_"..character,
    animfunc = "static",
    playing = true
  };
  
  local complete =
  {
    res = "assets/content/"..content.."/anm_character_"..character,
    animfunc = "complete",
    playing = true
  };
  
  local cell =
  {
    res = "assets/content/"..content.."/anm_character_cell",
    animfunc = "static_"..character,
    playing = true
  };
  
  local params = 
  {
    normal   = normal,
    complete = complete,
    cage     = cell,
    charm    = cell
  };
  
  ui.obj.Set( anm_name, params[ state ] );

end;
--------------------------------------------------------------------------------
-- fucntion .. () end;
--------------------------------------------------------------------------------
function private.Animate_Action ( character, state, func_after, is_clicked )

  local normal = 
  {
    action = tern( is_clicked )( "click", "action" ),
    static = "static"
  };
  
  local complete = 
  {
    action = "click_complete",
    static = "complete"
  };  
  
  local cell =
  {
    action = "action_"..character,
    static = "static_"..character
  };
  
  local animfunc =
  {
    normal   = normal,
    complete = complete,
    cage     = cell,
    charm    = cell
  };
  
  ui.anim.Play(
    "anm_map_characters_icon_"..character,
    animfunc[ state ].action,
    func_after,
    animfunc[ state ].static
  );

end;
--------------------------------------------------------------------------------
-- fucntion .. () end;
--------------------------------------------------------------------------------
function private.GetAnimData_QuestStart ( character, quest )

  local obj_name = "obj_map_characters_"..character;
  local anm_name = "anm_map_characters_icon_"..character;

  local states = 
  {
    normal = -- Character in normal state
    {
      name        = obj_name,
      func_action = function ( func_after )
      
        private.Indicator_Enable( character, true );
        private.Animate_Action( character, "normal" );
        
        if ( func_after ) then func_after(); end;
      
      end
    },
    cage = -- Character in cage
    {
      name        = obj_name,
      res         = "assets/content/map/effects/anm_appearance",
      coords      = { pos_x = 0, pos_y = 0 },
      func_dummy  = function ()
      
        private.Sound_Play( "metamorphose" );
        private.Animate_Action( character, "cage" );
      
      end,
      func_action = function ( func_after )
      
        private.Indicator_Enable( character, true );
        if ( func_after ) then func_after(); end;
      
      end
    },
    charm = -- Transform character to charmed state
    {
      name        = obj_name,
      skip_focus  = true,
      func_action = function ( func_after )
      
        local enemy     = private.GetMainEnemy( character );
        local obj_enemy = "anm_map_point_enemy_"..enemy;
        
        -- Enemy scaling hack
        local scales =
        {
          [ "trollking_stage_330" ] = 0.6
        }
        
        v_map.Focus( obj_enemy,
        {
          scale      = scales[ enemy ] or 1.0,
          func_after = function ()
          
            local obj_parent = ObjGetRelations( obj_enemy ).parent;
            ObjClone( obj_enemy, "_dummy", 1, obj_parent );

            ui.anim.Play( obj_enemy.."_dummy", "dummy", function ()
            
              ObjDelete( obj_enemy.."_dummy" );
              v_map.Focus( anm_name,
              {
                scale      = 1.0,
                func_after = function ()
                
                  private.SetVisualState( character, "charm" );
                  private.Sound_Play( "metamorphose" );
                  
                  ui.anim.Play( anm_name, "catch", function ()
                  
                    ui.anim.Play( anm_name, "static_"..character );
                    private.Indicator_Enable( character, true );
                    
                    if ( func_after ) then func_after(); end;
                    
                  end );

                end
              } );
              
            end );

            ui.anim.Play( obj_enemy, "action", function ()
              ui.anim.Play( obj_enemy, "static" );
            end );
            local section = v_sections.Section_NameCurrent();
             common.sound.Play( "assets/content/"..section.."/aud_boss_attack" );
          end
        } );
      
      end
    }
  }
  
  local state = private.GetVisualState_Active( character, quest.name );
  return states[ state ];

end;
--------------------------------------------------------------------------------
function private.GetAnimData_ItemReceive ( character, quest )

  local obj_name = "obj_map_characters_"..character;
  local anm_name = "anm_map_characters_icon_"..character;

  local states = 
  {
    normal = -- Character in normal state
    {
      name        = obj_name,
      func_action = function ( func_after )
      
        ui.anim.Play( anm_name, "action", func_after, "static" );
        
        --if ( func_after ) then func_after(); end;
      
      end
    },
    cage = -- Character in cage
    {
      name        = obj_name,
      func_action = function ( func_after )
      
        ui.anim.Play( anm_name, "action_"..character, func_after, "static_"..character );
        
        --if ( func_after ) then func_after(); end;
      
      end
    },
    charm = -- Character in charmed state
    {
      name        = obj_name,
      func_action = function ( func_after )
      
        ui.anim.Play( anm_name, "action_"..character, func_after, "static_"..character );
        
        --if ( func_after ) then func_after(); end;
      
      end
    }
  }
  
  local state = private.GetVisualState_Active( character, quest.name );
  return states[ state ];

end;
--------------------------------------------------------------------------------
function private.GetAnimData_QuestDone ( character, quest )

  local obj_name = "obj_map_characters_"..character;
  
  --local state = private.GetVisualState_Active( character, quest.name );
  local state = private.GetVisualState_Complete( character, quest.name );
  
  local free_anim =
  {
    name        = obj_name,
    res         = "assets/content/map/effects/anm_appearance_character",
    coords      = { pos_x = 0, pos_y = 0 },
    func_preset = function ()
    
      private.Indicator_Disable( character, true );
      private.Animate_Action( character, state );
    
    end,
    func_dummy  = function ()
    
      private.SetVisualState( character, "normal" );
      private.Animate_Action( character, "normal" );
      private.Sound_Play( "metamorphose" );
    
    end,
    func_action = function ( func_after )
      
      if ( func_after ) then func_after(); end;
    
    end
  }

  local states = 
  {
    normal = -- Character in normal state
    {
      name        = obj_name,
      func_action = function ( func_after )
      
        private.Indicator_Disable( character, true );
        private.Animate_Action( character, "normal", func_after );
        
      end
    },
    complete = -- Character in complete state
    {
      name        = obj_name,
      func_action = function ( func_after )
      
        private.Indicator_Disable( character, true );
        private.Animate_Action( character, "complete", func_after );
        
      end
    },
    cage  = free_anim, -- Transform cage to character
    charm = free_anim  -- Transform character from charmed to normal state
  };
  
  return states[ state ];

end;
--------------------------------------------------------------------------------
function public.Callback_AnimateAppear ( build_name )

  local character = m_characters.GetCharacterByBuild( build_name );

  if  ( character ) then
  
    local obj_name  = "obj_map_characters_"..character.name;
    local anm_name  = "anm_map_characters_icon_"..character.name;
    local anm_chest = "anm_map_chest_"..character.name;
  
    local anim  = {};
    anim.name   = obj_name;
    anim.res    = "assets/content/map/effects/anm_appearance";
    anim.coords = { pos_x = 0, pos_y = 0 };
    
    anim.func_preset = function ()
    
      ui.obj.Set( obj_name, { alp = 0.0 } );

      if ( ui.obj.Exists( anm_chest ) ) then

        ui.obj.Set( anm_chest, { alp = 0.0 } );

      end;

      private.SetUnlocked( character );
    
    end;
    
    anim.func_dummy = function ()
    
      private.Sound_Play( "lighting_app" );
      ui.anim.FadeIn(
        0.3,
        obj_name,
        nil,
        function ()

          if ( ui.obj.Exists( anm_chest ) ) then

            ui.anim.FadeIn(
              0.3,
              anm_chest,
              nil,
              function () ui.anim.Bounce( anm_chest ); end
            );

          end;

        end
      );
      
      local state = private.GetVisualState_Unlocked( character.name );
      private.Animate_Action( character.name, state );
    
    end;
    
    anim.func_action = function ( func_after )
      
      if ( func_after ) then func_after(); end;
    
    end;
    
    return v_map.Callback_Animate( anim );
    
  end;
  
  return nil;

end;
--------------------------------------------------------------------------------