-- name=v_mines
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "m_mines",
  "int_dialog_map_mine"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );
  
  private.signs    = {};
  private.progress = {};
  
  mediator.Subscribe(
    "Model_Mines_MineCooldownStop",
    private.Mine_OnCooldownFinished
  );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Init ()
  
  v_game.RegisterFeature( 
  { 
    entity = "v_mines", 
    feature = "gold_mine_fresh_player",
    func_save = private.Feature_SaveData
  } );
  
  local funcs = {
    func_get_stats = function ()
      return v_player.Stats_GetPlayerData();
    end,
    func_get_remain_time_speedup = function ()
      return v_potions.GetRemainingTime( "potion_mines_speed" );
    end;
  }
  
  m_mines.Setup( funcs );
  
  mediator.Subscribe(
    "Model_Potion_RunAction",
    private.Cooldown_Speedup_Start
  );
  
  mediator.Subscribe(
    "Model_Potion_CompleteAction",
    private.Cooldown_Speedup_Stop
  );
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Feature_SaveData ()
  
  v_features.Feature_SaveState( "gold_mine_fresh_player" );
  
end;
--------------------------------------------------------------------------------
function public.FeatureUpdated ()
  
  --
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Cooldown_Speedup_Start ()
  
  m_mines.SpeedUpTimer();
  
end;
--------------------------------------------------------------------------------
function private.Cooldown_Speedup_Stop ()
  
  m_mines.SpeedUpTimerFinished();
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.GetAllBySection ( section_name )

  local mines = m_mines.GetAllBySection( section_name );
  
  return mines;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
--[[
  Mine reaction format:
  {
    -- Basic data:
    name       = "build_name",
    func_popup = function ( func_after ) end,
    func_anim  = function ( func_after ) end,
    
    -- Only for reward drop:
    object          = "obj_build_name", -- receiver-object
    func_drop       = function () end,  -- on-drop function
    call_drop_first = true, -- true : func_drop on focus, func_anim on drop
                            -- false: func_anim on focus, func_drop on drop
  }
]]
--------------------------------------------------------------------------------
function public.GetRewardReaction ( stage, rewards )

  local reaction;
  
  if not ( rewards ) then return; end;
  
  -- Find worker
  local worker = nil;
  for i, reward in pairs( rewards ) do
    if ( reward.type == "worker" ) then
      worker = reward;
      break;
    end;
  end;
  
  local mine_name = public.Mine_GetNameByStage( stage );
  
  if not ( worker and mine_name ) then return; end;
  
  local mine = m_mines.Mine_Data( mine_name );
  
  if ( m_mines.IsMineUnavailable( mine_name ) ) then
    return;
  end;
  
  if ( mine.time ) and ( mine.time < 1 ) then return; end;
  
  reaction =
  {
    name            = mine.name,
    object          = "obj_map_mines_"..mine.name,
    call_drop_first = true,
    func_anim       = function ( func_after )
    
      ui.screen.Lock( true );
      
      ui.obj.Set( "spr_map_mines_icon_"..mine.name,
        {
          res      = ui.obj.Res( "mine", "base" ),
          animfunc = "state_"..math.ceil( mine.progress * 10 / mine.progress_max ),
          playing  = true,
          endtrig  = function ()
          
            ui.screen.Lock( false );
            if ( func_after ) then func_after(); end;
          
          end;
        }
      );
    
      private.Progress_Show( mine.name, mine.progress, mine.progress_max );
      
      if ( mine.progress == mine.progress_max ) then
        
          private.Indicator_Show( mine.name );
        
      end;
    
    end
  };
  
  -- Add popup for first seed
  if ( mine.progress == 1 ) then
  
    reaction.func_popup = function ( func_after )
    
      private.Popup_Collect( mine.name, func_after );
    
    end;
  
  end;
  
  return reaction;

end;
--------------------------------------------------------------------------------
function public.GetStageCompleteReaction ( stage )

  local reaction;
  
  local mine_name = m_mines.GetMineByUnlockStage( stage );
  
  if not ( mine_name ) then return; end;
  
  local mine_type =  m_mines.Mine_Type( mine_name );
  local mine_time = m_mines.GetMineTime( mine_type );
  
  if ( mine_time < 1 ) then return; end;
  
  if ( m_mines.IsMineUnavailable( mine_name ) ) then
    return;
  end;
  
  if ( mine_type == "gold" ) then
  
    reaction = 
    {
      name   = mine_name,
      object = "obj_map_mines_"..mine_name,
      call_drop_first = true,
      func_anim       = function ( func_after )
      
        private.Mine_AnimateUpgrade( mine_name );
        if ( func_after ) then func_after(); end;
      
      end
    };
  
  end;
  
  return reaction;

end;
--------------------------------------------------------------------------------
function public.Usable_Search ( resource_name )

  local section_name = v_sections.Section_NameCurrent();
  local usable_mines = m_mines.Usable_List( section_name );

  if ( usable_mines ) then

    for i = 1, #usable_mines do

      if ( not m_mines.IsMineUnavailable( usable_mines[ i ].name ) ) then
      
        local count = m_mines.Progress_GetCount( usable_mines[ i ].name ).count;

        if ( usable_mines[ i ].resource == resource_name ) and ( count > 5 ) then

          return usable_mines[ i ];

        end;
        
      end;

    end;

    return false;

  end;

  return false;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function public.Mine_Create ( mine )

  local obj_name = ( "obj_map_mines_%s"      ):format( mine.name );
  local btn_name = ( "btn_map_mines_%s"      ):format( mine.name );
  local sgn_name = ( "obj_map_mines_sign_%s" ):format( mine.name );

  ObjCreate( obj_name, "obj" );
  ObjSet( obj_name, { anim_tag = mine.name } );
  ObjAttach( obj_name, "anm_map_"..mine.section );

  ObjClone( "btn_map_mines", "_"..mine.name, 1, obj_name );
  
  local mine_type = m_mines.Mine_Type( mine.name );
  local mine_time = m_mines.GetMineTime( mine_type );
  
  if ( not m_mines.IsMineUnavailable( mine.name ) ) then
    
    local func_click = function () private.Mine_Tap( mine.name ); end
    
    ui.control.EventsSet( btn_name, { func_on_click = func_click } );
    ui.control.EventsSet( sgn_name, { func_on_click = func_click } );
  
  end;
  
  private.Mine_Set( mine.name, true );

end;
--------------------------------------------------------------------------------
function public.Mine_Data ( mine_name )
  
  local mine_data = m_mines.Mine_Data( mine_name );
  
  return mine_data;
  
end;
--------------------------------------------------------------------------------
function public.Mine_Type ( mine_name )
  
  local mine_type, mine_resource = m_mines.Mine_Type( mine_name );
  return mine_type, mine_resource;
  
end;
--------------------------------------------------------------------------------
function public.Mine_GetNameByStage ( stage_name )
  
  for _, mine in pairs( m_mines.ListAll() ) do

    if ( mine.progress and #mine.progress > 0 ) then

      for i = 1, #mine.progress do

        if ( mine.progress[ i ] == stage_name ) then

          return mine.name;

        end;

      end;

    end;

  end;

  return nil;
  
end;
--------------------------------------------------------------------------------
function public.Mine_TryUnlockByStage ( stage_name, reward_data )
  
  local mine_name = m_mines.Mine_GetNameByUnlock( stage_name );
  
  if ( not mine_name ) then
    
    mine_name = public.Mine_GetNameByStage( stage_name );

  end;
  
  if ( not mine_name ) then

    return false;

  end;
  
  local mine_data = m_mines.Mine_Data( mine_name );
  
  if ( mine_data.unlocked ) and
     ( not reward_data    ) then
    
    return false;
    
  end;
  
  if ( not mine_data.unlocked ) then
  
    local unlock_data  = m_mines.Mine_Unlock( mine_name );
    mine_data.progress = unlock_data.progress;
  
  end;

  if ( mine_data.type == "gold" ) then

    local mine_time = m_mines.GetMineTime( mine_data.type );
    
    if ( mine_time < 1 ) then
      return false;
    end;
    
    private.Mine_TrySetResource( mine_name, "gold" );

  else
  
    local new_progress = mine_data.progress;
    
    if ( reward_data ) then
        
      for _, reward in pairs( reward_data ) do

        if ( reward.name == "worker" ) then

          new_progress = new_progress + reward.count;

          break;

        end;

      end;
      
      if ( new_progress > mine_data.progress ) then
      
        m_mines.Mine_SetProgress(
          mine_name,
          mine_data.progress,
          new_progress
        );
        
        if ( new_progress == mine_data.progress_max ) then
        
          m_mines.Mine_SetStateWait( mine_name, mine_data.progress );
        
        end;
      
      end;
      
    end;

  end;
  
  return true;
  
end;
--------------------------------------------------------------------------------
function public.Mine_GetRandomCooldown ()
  
  return m_mines.Mine_GetRandomCooldown();

end;
--------------------------------------------------------------------------------
function public.Mine_GetStaticData ( mine_name )
  
  local  mine_data = m_mines.Mine_GetStaticData( mine_name );
  return mine_data

end;
--------------------------------------------------------------------------------
function public.Mine_GetSavedData ( mine_name )
  
  local  mine_data = m_mines.Mine_GetSavedData( mine_name );
  return mine_data

end;
--------------------------------------------------------------------------------
function public.Mine_SetProgress ( mine_name, old_progress, new_progress )
  
  local is_set = m_mines.Mine_SetProgress(
    mine_name, old_progress, new_progress
  );

  return is_set;

end;
--------------------------------------------------------------------------------
function public.Mine_SetStateWait ( mine_name )
  
  local  is_set = m_mines.Mine_SetStateWait( mine_name );
  return is_set;

end;
--------------------------------------------------------------------------------
function public.Mines_GetAllConstructed ()
  
  local  constructed = m_mines.Mines_GetAllConstructed();
  return constructed;
  
end;
--------------------------------------------------------------------------------
function public.IsMineUnavailable ( mine_name )
  
  local state = m_mines.IsMineUnavailable( mine_name );
  return state;
  
end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Mine_Set ( mine_name, instant )
  
  local anm_mine = "spr_map_mines_icon_"..mine_name;
  local mine = m_mines.Mine_Data( mine_name );
  mine.state = mine.state or "none";
  
  private.Progress_Hide( mine.name );
  private.Upgrade_Hide( mine.name );
  private.Sign_Hide( mine.name, instant );
  private.Indicator_Hide( mine.name, instant );
  
  if ( m_mines.IsMineUnavailable( mine_name ) ) then
    ui.obj.Set( anm_mine,
      {
        res      = ui.obj.Res( "mine", "base" ),
        animfunc = "static_"..mine.type,
        playing  = true 
      }
    );
    return;
  end;
  
  -- States:
  -- none   - collecting
  -- wait   - ready to build
  -- active - producing resources
  -- ready  - full of resources
  
  if ( mine.state == "none" or mine.state == "wait" ) then
  
   if ( mine.progress and mine.progress > 0 ) then
    
      ui.obj.Set( anm_mine,
        {
          res      = ui.obj.Res( "mine", "base" ),
          animfunc = "state_"..math.ceil( mine.progress * 10 / mine.progress_max ),
          playing  = true 
        }
      );
    
      private.Progress_Show( mine.name, mine.progress, mine.progress_max );
    
    else
    
      ui.obj.Set( anm_mine,
        {
          res      = ui.obj.Res( "mine", "base" ),
          animfunc = "static_"..mine.type,
          playing  = true 
        }
      );
    
    end;
    
    if ( mine.state == "wait" ) then
    
      private.Indicator_Show( mine.name, instant );
    
    end;
  
  elseif ( mine.state == "active" or mine.state == "ready" ) then
  
    local level    = mine.level or 1;
    local resource = mine.resource or "resource1";
    
    ui.obj.Set( anm_mine,
      {
        res        = ui.obj.Res( "mine", resource ),
        animfunc   = "upgrade_"..level.."_static",
        playing    = true,
        tt_counter = ""
      }
    );
    
    private.Sign_TryShow( mine.name );
    private.Sign_CreateTimer( mine.name );
  
    if ( mine.type ~= "gold" and level < 3 ) then
    
      private.Upgrade_Show( mine.name, resource );
    
    end;
  
  end;

end;
--------------------------------------------------------------------------------
function private.Mine_Tap ( mine_name )

  local collected = private.Mine_TryCollectResource( mine_name );
  local mine_data = m_mines.Mine_Data( mine_name );
  local obj_name  = "obj_map_mines_"..mine_name;
  
  mine_data.state = mine_data.state or "none";

  if ( collected ) then

    local func_OnFocus = function ()

      private.Sign_Hide( mine_name );
      private.Mine_AnimateGet( mine_data.name, mine_data.level );

      v_map.Reward_ShowCollected(
        mine_name,
        collected.count,
        collected.name
      );
      
    end;
    
    v_map.Focus( obj_name, { func_after = func_OnFocus } );

  else

    if ( mine_data.state == "none" ) then

      private.Popup_Collect( mine_name );

    elseif ( mine_data.state == "wait" ) then

      if ( mine_data.price > 0 ) then v_interface.panel.top.Lock(); end;

      v_map.Focus( obj_name,
        {
          func_after = function () private.DialogSelect_Show( mine_name ); end
        }
      );

    else

      private.Popup_Wait( mine_name );

    end;

  end;

end;
--------------------------------------------------------------------------------
function private.Mine_TryUpgrade ( mine_name )

  local mine_data = m_mines.Mine_Data( mine_name );
  local mine_type = m_mines.Mine_Type( mine_name );
  
  local is_upgradable, price, new_level =
    m_mines.Mine_IsUpgradable( mine_type, mine_data.level );
  
  if ( not is_upgradable ) then
    
    return false;
    
  end;
  
  if ( v_player.Gold_TryChangeCount( -price ) ) then
  
    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "premium_currency_spend",
      category    = "farm_upgrade",
      value       = -price,
      name        = mine_name
    } );

    return m_mines.Mine_Upgrade( mine_name, new_level );
  
  end;
  
  return false;
  
end;
--------------------------------------------------------------------------------
function private.Mine_TrySetResource ( mine_name, resource_name )

  local mine_data = m_mines.Mine_Data( mine_name );

  if ( mine_data.state ~= "wait" ) and
     ( mine_data.type  ~= "gold" ) then

    return false;

  end;

  if ( mine_data.price == 0                             ) or
     ( v_player.Gold_TryChangeCount( -mine_data.price ) ) then

    v_log.Kontagent_LogEvent( "economy", 
    {
      description = "premium_currency_spend",
      category    = "farm_build",
      value       = -mine_data.price,
      name        = mine_name
    } );

    m_mines.Mine_SetResource( mine_name, mine_data.type, resource_name );

    return true;

  end;

  return false;

end;
--------------------------------------------------------------------------------
function private.Mine_TryCollectResource ( mine_name )

  local progress = m_mines.Progress_GetCount( mine_name );

  if ( progress.count == 0 ) then

    return false;

  end;

  local mine_type = public.Mine_Type( mine_name );

  local func_GetProgress = 
  {
    gold = function () 

      local is_applied = v_player.Gold_TryChangeCount( progress.count, true );

      if ( is_applied ) then

        v_log.Kontagent_LogEvent( "economy", 
        {
          description = "premium_currency_collect",
          category    = "farm_gather",
          value       = progress.count,
          name        = mine_name
        } );
      
        m_mines.Mine_OnResourceCollected( mine_name, mine_type );

        return progress;
      
      end;
      
      return false; 

    end,
    resource = function ()
               
      local is_applied = v_player.Resource_TryChangeCount(
        progress.name,
        progress.count,
        true
      );
 
      if ( is_applied ) then

        v_log.Kontagent_LogEvent( "economy", 
        {
          description = "soft_currency_collect",
          category    = "farm_gather",
          value       = progress.count,
          name        = mine_name,
          type        = progress.name
        } );

        m_mines.Mine_OnResourceCollected( mine_name, mine_type );

        return progress;
      
      end;
    
      return false;

    end
  };

  return func_GetProgress[ mine_type ]();

end;
--------------------------------------------------------------------------------
function private.Mine_AnimateUpgrade ( mine_name, func_after )

  local mine_data = m_mines.Mine_Data( mine_name ); 
  local level     = mine_data.level or 1;
  local resource  = mine_data.resource or "resource1";
  local anm_name = "spr_map_mines_icon_"..mine_name;
  
   common.sound.Play( "unspell" );
  
  ui.obj.Set( anm_name,
    {
      visible    = true,
      res        = ui.obj.Res( "mine", resource ),
      tt_counter = ""
    }
  );
  
  -- Preanim
  if ( mine_data.type == "resource" ) then
  
    private.Upgrade_Hide( mine_name );
  
    if ( level == 1 ) then
    
      private.Indicator_Hide( mine_name, true );
      private.Progress_Hide( mine_name );
      
    end;
  
  end;
  
  if ( level == 1 ) then
  
    private.Sign_CreateTimer( mine_name );
  
  end;
  
  ui.anim.Play(
    anm_name,
    "upgrade_"..level.."_appear",
    function ()

      if ( mine_data.type == "resource" and level < 3 ) then
      
        private.Upgrade_Show( mine_name, resource );
        
      end;      

      private.Sign_TryShow( mine_name );
      
      if ( func_after ) then func_after(); end;

    end,
    "upgrade_"..level.."_static"
  );

end;
--------------------------------------------------------------------------------
function private.Mine_AnimateGet ( mine_name, mine_level, func_after )
  
  mine_level = mine_level or 1;
  
  ui.anim.Play(
    "spr_map_mines_icon_"..mine_name,
    "upgrade_"..mine_level.."_get",
    func_after,
    "upgrade_"..mine_level.."_static"
  );

end;
--------------------------------------------------------------------------------
function private.Mine_OnCooldownFinished ( mine_name )

  if ( private.signs[ mine_name ] ) then

    private.Sign_StartAnimation( mine_name );

  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Progress_Show ( mine_name, current, total )

  current = ui.text.FormatNumber( current );
  total   = ui.text.FormatNumber( total   );

  local obj_progress = ( "obj_map_mines_progress_%s" ):format( mine_name );
  local txt_progress = ( "txt_map_mines_progress_%s" ):format( mine_name );
  
  local set = {};
  
  ui.obj.Enable( obj_progress, true, set );
  ui.obj.Set( txt_progress, 
    {
      active  = true,
      visible = true,
      text    = ( "%s/%s" ):format( current, total )
    }, 
    set
  );
  
  ui.obj.Multiset( set );

end;
--------------------------------------------------------------------------------
function private.Progress_Hide ( mine_name )

  local obj_progress = ( "obj_map_mines_progress_%s" ):format( mine_name );
  ui.obj.Disable( obj_progress, true );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Indicator_Show ( mine_name, instant )

  local indicator = "anm_map_mines_indicator_"..mine_name;

  if not ( instant ) then
  
    ui.obj.Enable( indicator );
    ObjAnimate( indicator, "scale_xy", 0, 0, "", 
      { 
        0.0, 0, 0.0, 0.0,  
        0.3, 0, 1.0, 1.0
      }
    );
  
  else
  
    ui.obj.Enable( indicator );
  
  end;
  
end;
--------------------------------------------------------------------------------
function private.Indicator_Hide ( mine_name, instant )

  local indicator = "anm_map_mines_indicator_"..mine_name;
  
  if not ( instant ) then
  
    ObjAnimate( indicator, "scale_xy", 0, 0,
      function ()
        ui.obj.Disable( indicator );
      end, 
      { 
        0.0, 0, 0.0, 0.0,  
        0.3, 0, 1.0, 1.0
      }
    );
  
  else
  
    ui.obj.Disable( indicator );
  
  end;

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Upgrade_Show ( mine_name, resource )

  local button = "btn_map_mines_upgrade_"..mine_name;
  
  ui.button.Create( button,
    {
      title  = "str_int_dialog_map_mine_upgrade_accept",
      design = "upgrade_"..resource
    },
    function ()
      private.DialogUpgrade_Show( mine_name );
    end
  );
  
  ui.obj.Enable( button );

end;
--------------------------------------------------------------------------------
function private.Upgrade_Hide ( mine_name )

  ui.obj.Disable( "btn_map_mines_upgrade_"..mine_name );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Sign_CreateTimer ( mine_name )

  local tmr_name = "tmr_map_mines_"..mine_name;

  ui.obj.Create(
    tmr_name,
    "timer",
    "obj_map_mines_"..mine_name,
    {
      time    = 1.0,
      playing = true,
      endtrig = function()
      
        private.Sign_TryShow( mine_name );
        ui.obj.Set( tmr_name, { time = 1.0, playing = true } );

      end
    }
  );

end;
--------------------------------------------------------------------------------
function private.Sign_TryShow ( mine_name )

  local progress    = m_mines.Progress_GetCount( mine_name );
  local is_progress = progress.count >= 1;
  local is_changed  = ( private.progress[ mine_name ] ~= progress.count );

  if ( not private.signs[ mine_name ] ) and
     ( is_progress                    ) then
  
    private.signs[ mine_name ] = true;

    local multiset = {};

    ui.obj.Set(
      "spr_map_mines_sign_icon_"..mine_name,
      { 
        active   = true, 
        visible  = true, 
        res      = ui.obj.Res( "resource", "icon_"..progress.name, "" ), 
        alp      = 0.0 
      },
      multiset
    );

    ui.obj.Set(
      "spr_map_mines_icon_"..mine_name,
      {
        tt_counter = ui.text.FormatNumber( progress.count )
      },
      multiset
    );

    ui.obj.Multiset( multiset );

    ui.anim.Play( "anm_map_mines_sign_bg_"..mine_name, "resource_popup_open" );
    ui.anim.FadeIn( 0.3, "spr_map_mines_sign_icon_"..mine_name );
    
    if ( ( progress.count / progress.full ) > 0.99 ) then

      private.Sign_StartAnimation( mine_name );

    end;

  elseif ( is_progress ) and
         ( is_changed  ) then

    ui.obj.Set(
      "spr_map_mines_icon_"..mine_name,
      {
        tt_counter = ui.text.FormatNumber( progress.count )
      }
    );

  end;

  if ( is_changed ) then

    private.progress[ mine_name ] = progress.count;

  end;

end;
--------------------------------------------------------------------------------
function private.Sign_StartAnimation ( mine_name )

  ObjAnimate( "obj_map_mines_sign_"..mine_name, "scale_xy", 1, 0, "",
  { 
    0.0, 3, 1.0, 1.0,
    1.0, 3, 1.5, 1.5,
    2.0, 3, 1.0, 1.0
  } );

end;
--------------------------------------------------------------------------------
function private.Sign_Hide ( mine_name, instant )

  private.signs[ mine_name ] = false;

  if not ( instant ) then

    ui.anim.Play(
      "anm_map_mines_sign_bg_"..mine_name,
      "resource_popup_close",
      function ()
        private.Sign_Disable( mine_name );
      end
    );

    local alp = ObjGet( "spr_map_mines_sign_icon_"..mine_name ).alp;

    ObjAnimate( "spr_map_mines_sign_icon_"..mine_name, "alp", 0, 0, "",
    {
      0.0, 0, alp,
      0.3, 0, 0.0
    } );

  else

    private.Sign_Disable( mine_name );

  end;

end;
--------------------------------------------------------------------------------
function private.Sign_Disable ( mine_name )

  local obj_sign = "obj_map_mines_sign_"..mine_name;
  local spr_sign = "spr_map_mines_sign_icon_"..mine_name;

  ObjStopAnimate( obj_sign, "scale_xy" );

  local multiset = {};
  
  ui.obj.Disable( spr_sign, true, multiset );
  ui.obj.Set( obj_sign, { scale_x = 1, scale_y = 1 }, multiset );
  ui.obj.Set( "spr_map_mines_icon_"..mine_name, { tt_counter = "" }, multiset );
  ui.obj.Multiset( multiset );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.DialogUpgrade_Show ( mine_name )

  local mine_data    = m_mines.Mine_Data( mine_name );
  local upgrade_data = m_mines.Mine_UpgradeData( mine_name );

  local delegates = {

    Accept = function ()

      if ( private.Mine_TryUpgrade( mine_name ) ) then

        private.Mine_AnimateUpgrade( mine_name );

      end;

    end,

    Close = private.DialogUpgrade_Hide,

    ShowTopPanel = function ()

      v_interface.panel.top.Highlight( "gold" );
      v_interface.panel.top.Show();

    end,

    HideTopPanel = v_interface.panel.top.Hide

  };

  common.delegate.Set( "int_dialog_map_mine", delegates );

  v_map.Focus(
    "obj_map_mines_"..mine_name,
    {
      func_after = function () 

        v_interface.window.Show(
          "int_dialog_map_mine",
          function ()

            int_dialog_map_mine.Upgrade_Show(
              mine_data.resource,
              upgrade_data
            );

          end
        );

      end
    }
  );

end;
--------------------------------------------------------------------------------
function private.DialogUpgrade_Hide ()

  local func_close = function ()

    v_interface.window.Hide( "int_dialog_map_mine" );

  end;

  int_dialog_map_mine.Upgrade_Hide( func_close );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.DialogSelect_Show ( mine_name )

  local mine_data    = m_mines.Mine_Data( mine_name ); 
  local mine_price   = mine_data.price;
  local produce_data = mine_data.produce;

  local resource_types = { "gold" };

  if ( mine_data.type == "resource" ) then
  
    resource_types = m_mines.GetResourcesTypes( mine_name );
    
  end;

  local delegates = {

    Accept = function ( resource )

      if ( resource ) then

        if ( private.Mine_TrySetResource( mine_name, resource ) ) then

          private.Mine_AnimateUpgrade( mine_name );

        end;

      end;

    end,

    Close = private.DialogSelect_Hide

  };

  common.delegate.Set( "int_dialog_map_mine", delegates );

  v_interface.window.Show(
    "int_dialog_map_mine",
    function ()

      int_dialog_map_mine.Select_Show(
        mine_data.type,
        resource_types,
        mine_price,
        produce_data
      );

    end
  );

end;
--------------------------------------------------------------------------------
function private.DialogSelect_Hide ()

  local func_close = function ()

    v_interface.window.Hide( "int_dialog_map_mine" );

  end;

  int_dialog_map_mine.Select_Hide( func_close );

end;
--------------------------------------------------------------------------------
-- function ... ()
--------------------------------------------------------------------------------
function private.Popup_Collect ( mine_name, func_after )

  local text      = "str_int_popup_cloud_mines_"..mine_name;
  local indicator = "obj_map_mines_indicator_"..mine_name;
  
  local funcs = { func_preclose = func_after };
  
  v_map.ShowObjectPopup( "monolog", indicator, text, nil, nil, funcs );

end;
--------------------------------------------------------------------------------
function private.Popup_Wait ( mine_name )

  local text      = "str_int_popup_cloud_mines_started";
  local indicator = "obj_map_mines_indicator_"..mine_name;

  if ( m_mines.Mine_Type( mine_name ) == "gold" ) then

    text = text.."_gold";

  end;

  v_map.ShowObjectPopup( "monolog", indicator, text );

end;
--------------------------------------------------------------------------------