-- name=v_screen
--------------------------------------------------------------------------------
-- function Modules ()
--------------------------------------------------------------------------------
local _modules = {
  "int_screen"
};
--------------------------------------------------------------------------------
function public.__GetDependencies ()

  return _modules;

end;

--------------------------------------------------------------------------------
-- Graphical content switcher controller.
-- Used for attach and detach root graphic objects
-- to 'ng_level' object. Only one screen can be active at one time.
-- Screen switching cause calling of some screen-related callbacks in specified
-- order.
-- @module v_screen
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Screen switch notifications. Each notification pass screen name as parameter.
--------------------------------------------------------------------------------
public.Notifications = {
  ScreenClose      = "Screen_Close",      -- notify before func_close.
  ScreenAfterClose = "Screen_AfterClose", -- notify before func_after_close.
  ScreenClosed     = "Screen_Closed",     -- notify after func_after_close.
  ScreenPreOpen    = "Screen_PreOpen",    -- notify before func_pre_open.
  ScreenOpen       = "Screen_Open",       -- notify before func_open.
  ScreenOpened     = "Screen_Opened"      -- notify after func_open.
};

public.LOADER = "scr_loader";
public.MAP    = "scr_map";
public.STAGE  = "scr_stage";
--------------------------------------------------------------------------------
-- Set module to initial state, load child modules.
--------------------------------------------------------------------------------
function public.__InitModule ()

  common.mod.Load( _modules );

  private.screens        = {};
  private.screen_current = nil;
  private.is_opening     = false;
  private.is_closing     = false;
  
end;

--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Setup screen switch callbacks.
-- Each callback is optional and has a 'func_after' argument.
--    <ul>
--    <li> func_pre_open    - background open operations, splash visible
--    <li> func_open        - visual open operations, splash hidden
--    <li> func_close       - visual close operations, splash hidden
--    <li> func_after_close - background close operations, splash visible
--    </ul>
-- @param scr_obj_name screen object name, must be created before.
-- @param callbacks callbacks table.
--------------------------------------------------------------------------------
function public.Setup ( scr_obj_name, callbacks )
  
  private.screens[ scr_obj_name ] = callbacks;
  
end;

--------------------------------------------------------------------------------
-- Set initial screen switch callbacks.
-- @param scr_obj_name screen object name, must be created before.
-- @param callbacks callbacks table, see public.Setup.
--------------------------------------------------------------------------------
function public.SetInitial ( scr_obj_name, callbacks )
  
  public.Setup( scr_obj_name, callbacks );
  private.screen_current = scr_obj_name;
  
end;

--------------------------------------------------------------------------------
-- Switch to specified screen, detach old, attach new.
-- @param scr_obj_name screen object name.
-- @param func_after_switch after switch callback, optional.
--------------------------------------------------------------------------------
function public.Switch ( scr_obj_name, func_after_switch )
  
  if not ( scr_obj_name ) then
    
    DbgTrace( "[ ERROR ] screen.Switch: Unknown screen: " .. scr_obj_name );
    return;
    
  end;
  
  local function func_switch_screen ( func_after )
  
    local screen_old = private.screen_current;
    local screen_new = scr_obj_name;
  
    if ( screen_old ) then
      
      int_screen.Detach( screen_old );
      mediator.Notify( public.Notifications.ScreenClosed, screen_old );
      
      private.is_closing = false;

    end;
    
    int_screen.Attach( screen_new );
    private.screen_current = screen_new;
    common.Save( true );
    
    func_after();
    
  end;
  
  local function func_end_switch ()

    private.is_opening = false;

    mediator.Notify( public.Notifications.ScreenOpened, scr_obj_name );
  
    if ( func_after_switch ) then func_after_switch() end;
  
  end;
  
  common.RunCallbackList(
    { 
      private.Screen_Close,
      private.ShowSplash,
      private.Screen_AfterClose,
      func_switch_screen,
      private.Screen_PreOpen,
      private.HideSplash,
      private.Screen_Open,
      func_end_switch
    }
  );
  
end;

--------------------------------------------------------------------------------
-- Get current screen name.
-- @return current screen name.
--------------------------------------------------------------------------------
function public.GetCurrentName ()
  
  return private.screen_current;
  
end;

--------------------------------------------------------------------------------
function public.IsOpening ()
  
  return private.is_opening;
  
end;

--------------------------------------------------------------------------------
function public.IsClosing ()
  
  return private.is_closing;
  
end;

--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Send notification and run screen close callback.
-- @param func_after after screen close callback.
--------------------------------------------------------------------------------
function private.Screen_Close ( func_after )
  
  if not ( private.screen_current ) then
    
    func_after();
    
  else 
  
    private.is_closing = true;
  
    local screen = private.screen_current;
    mediator.Notify( public.Notifications.ScreenClose, screen );
    
    local func_close = private.screens[ screen ].func_close or private.Empty;
    func_close( func_after );
    
  end;
  
end;

--------------------------------------------------------------------------------
-- Send notification and run screen after close callback.
-- @param func_after after screen after close callback.
--------------------------------------------------------------------------------
function private.Screen_AfterClose ( func_after )
  
  if not ( private.screen_current ) then
    
    func_after();
    
  else 
    
    local screen = private.screen_current;
    mediator.Notify( public.Notifications.ScreenAfterClose, screen );
    
    local func_after_close =    private.screens[ screen ].func_after_close 
                             or private.Empty;
    
    func_after_close( func_after );
    
  end;
  
end;

--------------------------------------------------------------------------------
-- Send notification and run screen preopen callback.
-- @param func_after after screen preopen callback.
--------------------------------------------------------------------------------
function private.Screen_PreOpen ( func_after )
  
  private.is_opening = true;

  local screen = private.screen_current;
  mediator.Notify( public.Notifications.ScreenPreOpen, screen );
  
  local func_pre_open =    private.screens[ screen ].func_pre_open
                        or private.Empty;

  func_pre_open( func_after );
  
end;

--------------------------------------------------------------------------------
-- Send notification and run screen open callback.
-- @param func_after after screen open callback.
--------------------------------------------------------------------------------
function private.Screen_Open ( func_after )
  
  local screen = private.screen_current;
  mediator.Notify( public.Notifications.ScreenOpen, screen );
  
  local func_open     =    private.screens[ screen ].func_open
                        or private.Empty
  
  func_open( func_after );
  
end;

--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Non exists callback replacement. Just call passed callback. Used for put to
-- common.RunCallbackList sequence.
-- @param func_after calling function.
--------------------------------------------------------------------------------
function private.Empty ( func_after )
  
  func_after();
  
end;

--------------------------------------------------------------------------------
-- function ... () end
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- Show splash, shade screen object and call func_after callback.
-- @param func_after calling function.
--------------------------------------------------------------------------------
function private.ShowSplash ( func_after )
  
  if ( private.screen_current ) then
  
    int_screen.InputOff( private.screen_current );
    int_screen.ShadeOn( private.screen_current );
    
    v_interface.Loading_Show( true, func_after );
    
  else
    
    v_interface.Loading_Show( false, func_after );
    
  end;
  
end;

--------------------------------------------------------------------------------
-- Hide splash, unshade screen object and call func_after callback.
-- @param func_after calling function.
--------------------------------------------------------------------------------
function private.HideSplash ( func_after )

  int_screen.ShadeOff( private.screen_current );

  local function func_after_splash ()

    int_screen.InputOn( private.screen_current );
    if ( func_after ) then func_after(); end;

  end;
  
  v_interface.Loading_Hide( true, func_after_splash );
  
end;

--------------------------------------------------------------------------------